"""
Defines the unit tests for the :mod:`colour.characterisation.correction`
module.
"""

from __future__ import annotations

import numpy as np
import platform
import unittest
from itertools import permutations
from numpy.linalg import LinAlgError

from colour.characterisation.correction import (
    matrix_augmented_Cheung2004,
    polynomial_expansion_Finlayson2015,
    polynomial_expansion_Vandermonde,
    matrix_colour_correction_Cheung2004,
    matrix_colour_correction_Finlayson2015,
    matrix_colour_correction_Vandermonde,
    colour_correction_Cheung2004,
    colour_correction_Finlayson2015,
    colour_correction_Vandermonde,
)
from colour.hints import NDArray
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "MATRIX_TEST",
    "MATRIX_REFERENCE",
    "TestMatrixAugmentedCheung2004",
    "TestPolynomialExpansionFinlayson2015",
    "TestPolynomialExpansionVandermonde",
    "TestMatrixColourCorrectionCheung2004",
    "TestMatrixColourCorrectionFinlayson2015",
    "TestMatrixColourCorrectionVandermonde",
    "TestColourCorrectionCheung2004",
    "TestColourCorrectionFinlayson2015",
    "TestColourCorrectionVandermonde",
]

MATRIX_TEST: NDArray = np.array(
    [
        [0.17224810, 0.09170660, 0.06416938],
        [0.49189645, 0.27802050, 0.21923399],
        [0.10999751, 0.18658946, 0.29938611],
        [0.11666120, 0.14327905, 0.05713804],
        [0.18988879, 0.18227649, 0.36056247],
        [0.12501329, 0.42223442, 0.37027445],
        [0.64785606, 0.22396782, 0.03365194],
        [0.06761093, 0.11076896, 0.39779139],
        [0.49101797, 0.09448929, 0.11623839],
        [0.11622386, 0.04425753, 0.14469986],
        [0.36867946, 0.44545230, 0.06028681],
        [0.61632937, 0.32323906, 0.02437089],
        [0.03016472, 0.06153243, 0.29014596],
        [0.11103655, 0.30553067, 0.08149137],
        [0.41162190, 0.05816656, 0.04845934],
        [0.73339206, 0.53075188, 0.02475212],
        [0.47347718, 0.08834792, 0.30310315],
        [0.00000000, 0.25187016, 0.35062450],
        [0.76809639, 0.78486240, 0.77808297],
        [0.53822392, 0.54307997, 0.54710883],
        [0.35458526, 0.35318419, 0.35524431],
        [0.17976704, 0.18000531, 0.17991488],
        [0.09351417, 0.09510603, 0.09675027],
        [0.03405071, 0.03295077, 0.03702047],
    ]
)

MATRIX_REFERENCE: NDArray = np.array(
    [
        [0.15579559, 0.09715755, 0.07514556],
        [0.39113140, 0.25943419, 0.21266708],
        [0.12824821, 0.18463570, 0.31508023],
        [0.12028974, 0.13455659, 0.07408400],
        [0.19368988, 0.21158946, 0.37955964],
        [0.19957424, 0.36085439, 0.40678123],
        [0.48896605, 0.20691688, 0.05816533],
        [0.09775522, 0.16710693, 0.47147724],
        [0.39358649, 0.12233400, 0.10526425],
        [0.10780332, 0.07258529, 0.16151473],
        [0.27502671, 0.34705454, 0.09728099],
        [0.43980441, 0.26880559, 0.05430533],
        [0.05887212, 0.11126272, 0.38552469],
        [0.12705825, 0.25787860, 0.13566464],
        [0.35612929, 0.07933258, 0.05118732],
        [0.48131976, 0.42082843, 0.07120612],
        [0.34665585, 0.15170714, 0.24969804],
        [0.08261116, 0.24588716, 0.48707733],
        [0.66054904, 0.65941137, 0.66376412],
        [0.48051509, 0.47870296, 0.48230082],
        [0.33045354, 0.32904184, 0.33228886],
        [0.18001305, 0.17978567, 0.18004416],
        [0.10283975, 0.10424680, 0.10384975],
        [0.04742204, 0.04772203, 0.04914226],
    ]
)


class TestMatrixAugmentedCheung2004(unittest.TestCase):
    """
    Define :func:`colour.characterisation.correction.\
matrix_augmented_Cheung2004` definition unit tests methods.
    """

    def test_matrix_augmented_Cheung2004(self):
        """
        Test :func:`colour.characterisation.correction.\
matrix_augmented_Cheung2004` definition.
        """

        RGB = np.array([0.17224810, 0.09170660, 0.06416938])

        polynomials = [
            np.array([0.17224810, 0.09170660, 0.06416938]),
            np.array(
                [0.17224810, 0.09170660, 0.06416938, 0.00101364, 1.00000000]
            ),
            np.array(
                [
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    0.01579629,
                    0.01105305,
                    0.00588476,
                    1.00000000,
                ]
            ),
            np.array(
                [
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    0.01579629,
                    0.01105305,
                    0.00588476,
                    0.00101364,
                    1.00000000,
                ]
            ),
            np.array(
                [
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    0.01579629,
                    0.01105305,
                    0.00588476,
                    0.02966941,
                    0.00841010,
                    0.00411771,
                    1.00000000,
                ]
            ),
            np.array(
                [
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    0.01579629,
                    0.01105305,
                    0.00588476,
                    0.02966941,
                    0.00841010,
                    0.00411771,
                    0.00101364,
                    1.00000000,
                ]
            ),
            np.array(
                [
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    0.01579629,
                    0.01105305,
                    0.00588476,
                    0.02966941,
                    0.00841010,
                    0.00411771,
                    0.00101364,
                    0.00511050,
                    0.00077126,
                    0.00026423,
                    1.00000000,
                ]
            ),
            np.array(
                [
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    0.01579629,
                    0.01105305,
                    0.00588476,
                    0.02966941,
                    0.00841010,
                    0.00411771,
                    0.00101364,
                    0.00272088,
                    0.00053967,
                    0.00070927,
                    0.00511050,
                    0.00077126,
                    0.00026423,
                ]
            ),
            np.array(
                [
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    0.01579629,
                    0.01105305,
                    0.00588476,
                    0.02966941,
                    0.00841010,
                    0.00411771,
                    0.00101364,
                    0.00272088,
                    0.00053967,
                    0.00070927,
                    0.00511050,
                    0.00077126,
                    0.00026423,
                    1.00000000,
                ]
            ),
            np.array(
                [
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    0.01579629,
                    0.01105305,
                    0.00588476,
                    0.02966941,
                    0.00841010,
                    0.00411771,
                    0.00101364,
                    0.00272088,
                    0.00053967,
                    0.00070927,
                    0.00190387,
                    0.00144862,
                    0.00037762,
                    0.00511050,
                    0.00077126,
                    0.00026423,
                ]
            ),
            np.array(
                [
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    0.01579629,
                    0.01105305,
                    0.00588476,
                    0.02966941,
                    0.00841010,
                    0.00411771,
                    0.00101364,
                    0.00272088,
                    0.00053967,
                    0.00070927,
                    0.00190387,
                    0.00144862,
                    0.00037762,
                    0.00511050,
                    0.00077126,
                    0.00026423,
                    1.00000000,
                ]
            ),
            np.array(
                [
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    0.01579629,
                    0.01105305,
                    0.00588476,
                    0.02966941,
                    0.00841010,
                    0.00411771,
                    0.00101364,
                    0.00272088,
                    0.00053967,
                    0.00070927,
                    0.00190387,
                    0.00144862,
                    0.00037762,
                    0.00511050,
                    0.00077126,
                    0.00026423,
                    0.00017460,
                    0.00009296,
                    0.00006504,
                ]
            ),
        ]

        for i, terms in enumerate(
            [3, 5, 7, 8, 10, 11, 14, 16, 17, 19, 20, 22]
        ):
            np.testing.assert_almost_equal(
                matrix_augmented_Cheung2004(RGB, terms),
                polynomials[i],
                decimal=7,
            )

    def test_raise_exception_matrix_augmented_Cheung2004(self):
        """
        Test :func:`colour.characterisation.correction.\
matrix_augmented_Cheung2004` definition raised exception.
        """

        self.assertRaises(
            ValueError,
            matrix_augmented_Cheung2004,
            np.array([0.17224810, 0.09170660, 0.06416938]),
            4,
        )

    @ignore_numpy_errors
    def test_nan_matrix_augmented_Cheung2004(self):
        """
        Test :func:`colour.characterisation.correction.\
matrix_augmented_Cheung2004` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            matrix_augmented_Cheung2004(case)


class TestPolynomialExpansionFinlayson2015(unittest.TestCase):
    """
    Define :func:`colour.characterisation.correction.\
polynomial_expansion_Finlayson2015` definition unit tests methods.
    """

    def test_polynomial_expansion_Finlayson2015(self):
        """
        Test :func:`colour.characterisation.correction.\
polynomial_expansion_Finlayson2015` definition.
        """

        RGB = np.array([0.17224810, 0.09170660, 0.06416938])

        polynomials = [
            [
                np.array([0.17224810, 0.09170660, 0.06416938]),
                np.array([0.17224810, 0.09170660, 0.06416938]),
            ],
            [
                np.array(
                    [
                        0.17224810,
                        0.09170660,
                        0.06416938,
                        0.02966941,
                        0.00841010,
                        0.00411771,
                        0.01579629,
                        0.00588476,
                        0.01105305,
                    ]
                ),
                np.array(
                    [
                        0.17224810,
                        0.09170660,
                        0.06416938,
                        0.12568328,
                        0.07671216,
                        0.10513350,
                    ]
                ),
            ],
            [
                np.array(
                    [
                        0.17224810,
                        0.09170660,
                        0.06416938,
                        0.02966941,
                        0.00841010,
                        0.00411771,
                        0.01579629,
                        0.00588476,
                        0.01105305,
                        0.00511050,
                        0.00077126,
                        0.00026423,
                        0.00144862,
                        0.00037762,
                        0.00070927,
                        0.00272088,
                        0.00053967,
                        0.00190387,
                        0.00101364,
                    ]
                ),
                np.array(
                    [
                        0.17224810,
                        0.09170660,
                        0.06416938,
                        0.12568328,
                        0.07671216,
                        0.10513350,
                        0.11314930,
                        0.07228010,
                        0.08918053,
                        0.13960570,
                        0.08141598,
                        0.12394021,
                        0.10045255,
                    ]
                ),
            ],
            [
                np.array(
                    [
                        0.17224810,
                        0.09170660,
                        0.06416938,
                        0.02966941,
                        0.00841010,
                        0.00411771,
                        0.01579629,
                        0.00588476,
                        0.01105305,
                        0.00511050,
                        0.00077126,
                        0.00026423,
                        0.00144862,
                        0.00037762,
                        0.00070927,
                        0.00272088,
                        0.00053967,
                        0.00190387,
                        0.00101364,
                        0.00088027,
                        0.00007073,
                        0.00001696,
                        0.00046867,
                        0.00032794,
                        0.00013285,
                        0.00004949,
                        0.00004551,
                        0.00002423,
                        0.00024952,
                        0.00003463,
                        0.00012217,
                        0.00017460,
                        0.00009296,
                        0.00006504,
                    ]
                ),
                np.array(
                    [
                        0.17224810,
                        0.09170660,
                        0.06416938,
                        0.12568328,
                        0.07671216,
                        0.10513350,
                        0.11314930,
                        0.07228010,
                        0.08918053,
                        0.13960570,
                        0.08141598,
                        0.12394021,
                        0.10045255,
                        0.14713499,
                        0.13456986,
                        0.10735915,
                        0.08387498,
                        0.08213618,
                        0.07016104,
                        0.11495009,
                        0.09819082,
                        0.08980545,
                    ]
                ),
            ],
        ]

        for i in range(4):
            np.testing.assert_almost_equal(
                polynomial_expansion_Finlayson2015(RGB, i + 1, False),
                polynomials[i][0],
                decimal=7,
            )
            np.testing.assert_almost_equal(
                polynomial_expansion_Finlayson2015(RGB, i + 1, True),
                polynomials[i][1],
                decimal=7,
            )

    def test_raise_exception_polynomial_expansion_Finlayson2015(self):
        """
        Test :func:`colour.characterisation.correction.\
polynomial_expansion_Finlayson2015` definition raised exception.
        """

        self.assertRaises(
            ValueError,
            polynomial_expansion_Finlayson2015,
            np.array([0.17224810, 0.09170660, 0.06416938]),
            5,
        )

    @ignore_numpy_errors
    def test_nan_polynomial_expansion_Finlayson2015(self):
        """
        Test :func:`colour.characterisation.correction.\
polynomial_expansion_Finlayson2015` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            polynomial_expansion_Finlayson2015(case)


class TestPolynomialExpansionVandermonde(unittest.TestCase):
    """
    Define :func:`colour.characterisation.correction.\
polynomial_expansion_Vandermonde` definition unit tests methods.
    """

    def test_polynomial_expansion_Vandermonde(self):
        """
        Test :func:`colour.characterisation.correction.\
polynomial_expansion_Vandermonde` definition.
        """

        RGB = np.array([0.17224810, 0.09170660, 0.06416938])

        polynomials = [
            np.array([0.17224810, 0.09170660, 0.06416938, 1.00000000]),
            np.array(
                [
                    0.02966941,
                    0.00841010,
                    0.00411771,
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    1.00000000,
                ]
            ),
            np.array(
                [
                    0.00511050,
                    0.00077126,
                    0.00026423,
                    0.02966941,
                    0.00841010,
                    0.00411771,
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    1.00000000,
                ]
            ),
            np.array(
                [
                    0.00088027,
                    0.00007073,
                    0.00001696,
                    0.00511050,
                    0.00077126,
                    0.00026423,
                    0.02966941,
                    0.00841010,
                    0.00411771,
                    0.17224810,
                    0.09170660,
                    0.06416938,
                    1.00000000,
                ]
            ),
        ]

        for i in range(4):
            np.testing.assert_almost_equal(
                polynomial_expansion_Vandermonde(RGB, i + 1),
                polynomials[i],
                decimal=7,
            )

    @ignore_numpy_errors
    def test_nan_polynomial_expansion_Vandermonde(self):
        """
        Test :func:`colour.characterisation.correction.\
polynomial_expansion_Vandermonde` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            polynomial_expansion_Vandermonde(case)


class TestMatrixColourCorrectionCheung2004(unittest.TestCase):
    """
    Define :func:`colour.characterisation.correction.\
matrix_colour_correction_Cheung2004` definition unit tests methods.
    """

    def test_matrix_colour_correction_Cheung2004(self):
        """
        Test :func:`colour.characterisation.correction.\
matrix_colour_correction_Cheung2004` definition.
        """

        np.testing.assert_almost_equal(
            matrix_colour_correction_Cheung2004(MATRIX_TEST, MATRIX_REFERENCE),
            np.array(
                [
                    [0.69822661, 0.03071629, 0.16210422],
                    [0.06893498, 0.67579611, 0.16430385],
                    [-0.06314956, 0.09212471, 0.97134152],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            matrix_colour_correction_Cheung2004(
                MATRIX_TEST, MATRIX_REFERENCE, terms=7
            ),
            np.array(
                [
                    [
                        0.80512769,
                        0.04001012,
                        -0.01255261,
                        -0.41056170,
                        -0.28052094,
                        0.68417697,
                        0.02251728,
                    ],
                    [
                        0.03270288,
                        0.71452384,
                        0.17581905,
                        -0.00897913,
                        0.04900199,
                        -0.17162742,
                        0.01688472,
                    ],
                    [
                        -0.03973098,
                        -0.07164767,
                        1.16401636,
                        0.29017859,
                        -0.88909018,
                        0.26675507,
                        0.02345109,
                    ],
                ]
            ),
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_matrix_colour_correction_Cheung2004(self):  # pragma: no cover
        """
        Test :func:`colour.characterisation.correction.\
        matrix_colour_correction_Cheung2004` definition nan support.
        """

        # NOTE: Hangs on "Linux".
        if platform.system() == "Linux":
            return

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = list(set(permutations(cases * 3, r=3)))[0:4]
        for case in cases:
            try:
                matrix_colour_correction_Cheung2004(
                    np.vstack([case, case, case]),
                    np.transpose(np.vstack([case, case, case])),
                )
            except LinAlgError:
                pass


class TestMatrixColourCorrectionFinlayson2015(unittest.TestCase):
    """
    Define :func:`colour.characterisation.correction.\
matrix_colour_correction_Finlayson2015` definition unit tests methods.
    """

    def test_matrix_colour_correction_Finlayson2015(self):
        """
        Test :func:`colour.characterisation.correction.\
matrix_colour_correction_Finlayson2015` definition.
        """

        np.testing.assert_almost_equal(
            matrix_colour_correction_Finlayson2015(
                MATRIX_TEST, MATRIX_REFERENCE
            ),
            np.array(
                [
                    [0.69822661, 0.03071629, 0.16210422],
                    [0.06893498, 0.67579611, 0.16430385],
                    [-0.06314956, 0.09212471, 0.97134152],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            matrix_colour_correction_Finlayson2015(
                MATRIX_TEST, MATRIX_REFERENCE, degree=3
            ),
            np.array(
                [
                    [
                        2.87796213,
                        9.85720054,
                        2.99863978,
                        76.97227806,
                        73.73571500,
                        -49.37563169,
                        -48.70879206,
                        -47.53280959,
                        29.88241815,
                        -39.82871801,
                        -37.11388282,
                        23.30393209,
                        3.81579802,
                    ],
                    [
                        -0.78448243,
                        5.63631335,
                        0.95306110,
                        14.19762287,
                        20.60124427,
                        -18.05512861,
                        -14.52994195,
                        -13.10606336,
                        10.53666341,
                        -3.63132534,
                        -12.49672335,
                        8.17401039,
                        3.37995231,
                    ],
                    [
                        -2.39092600,
                        10.57193455,
                        4.16361285,
                        23.41748866,
                        58.26902059,
                        -39.39669827,
                        -26.63805785,
                        -35.98397757,
                        21.25508558,
                        -4.12726077,
                        -34.31995017,
                        18.72796247,
                        7.33531009,
                    ],
                ]
            ),
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_matrix_colour_correction_Finlayson2015(
        self,
    ):  # pragma: no cover
        """
        Test :func:`colour.characterisation.correction.\
        matrix_colour_correction_Finlayson2015` definition nan support.
        """

        # NOTE: Hangs on "Linux".
        if platform.system() == "Linux":
            return

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = list(set(permutations(cases * 3, r=3)))[0:4]
        for case in cases:
            try:
                matrix_colour_correction_Finlayson2015(
                    np.vstack([case, case, case]),
                    np.transpose(np.vstack([case, case, case])),
                )
            except LinAlgError:
                pass


class TestMatrixColourCorrectionVandermonde(unittest.TestCase):
    """
    Define :func:`colour.characterisation.correction.\
matrix_colour_correction_Vandermonde` definition unit tests methods.
    """

    def test_matrix_colour_correction_Vandermonde(self):
        """
        Test :func:`colour.characterisation.correction.\
matrix_colour_correction_Vandermonde` definition.
        """

        np.testing.assert_almost_equal(
            matrix_colour_correction_Vandermonde(
                MATRIX_TEST, MATRIX_REFERENCE
            ),
            np.array(
                [
                    [0.66770040, 0.02514036, 0.12745797, 0.02485425],
                    [0.03155494, 0.66896825, 0.12187874, 0.03043460],
                    [-0.14502258, 0.07716975, 0.87841836, 0.06666049],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            matrix_colour_correction_Vandermonde(
                MATRIX_TEST, MATRIX_REFERENCE, degree=3
            ),
            np.array(
                [
                    [
                        -0.04328223,
                        -1.87886146,
                        1.83369170,
                        -0.10798116,
                        1.06608177,
                        -0.87495813,
                        0.75525839,
                        -0.08558123,
                        0.15919076,
                        0.02404598,
                    ],
                    [
                        0.00998152,
                        0.44525275,
                        -0.53192490,
                        0.00904507,
                        -0.41034458,
                        0.36173334,
                        0.02904178,
                        0.78362950,
                        0.07894900,
                        0.01986479,
                    ],
                    [
                        -1.66921744,
                        3.62954420,
                        -2.96789849,
                        2.31451409,
                        -3.10767297,
                        1.85975390,
                        -0.98795093,
                        0.85962796,
                        0.63591240,
                        0.07302317,
                    ],
                ]
            ),
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_matrix_colour_correction_Vandermonde(
        self,
    ):  # pragma: no cover
        """
        Test :func:`colour.characterisation.correction.\
        matrix_colour_correction_Vandermonde` definition nan support.
        """

        # NOTE: Hangs on "Linux".
        if platform.system() == "Linux":
            return

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = list(set(permutations(cases * 3, r=3)))[0:4]
        for case in cases:
            try:
                matrix_colour_correction_Vandermonde(
                    np.vstack([case, case, case]),
                    np.transpose(np.vstack([case, case, case])),
                )
            except LinAlgError:
                pass


class TestColourCorrectionCheung2004(unittest.TestCase):
    """
    Define :func:`colour.characterisation.correction.\
colour_correction_Cheung2004` definition unit tests methods.
    """

    def test_colour_correction_Cheung2004(self):
        """
        Test :func:`colour.characterisation.correction.\
colour_correction_Cheung2004` definition.
        """

        RGB = np.array([0.17224810, 0.09170660, 0.06416938])

        np.testing.assert_almost_equal(
            colour_correction_Cheung2004(RGB, MATRIX_TEST, MATRIX_REFERENCE),
            np.array([0.13348722, 0.08439216, 0.05990144]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            colour_correction_Cheung2004(
                RGB, MATRIX_TEST, MATRIX_REFERENCE, terms=7
            ),
            np.array([0.15850295, 0.09871628, 0.08105752]),
            decimal=7,
        )

    def test_n_dimensional_colour_correction_Cheung2004(self):
        """
        Test :func:`colour.characterisation.correction.\
colour_correction_Cheung2004` definition n-dimensional support.
        """

        RGB = np.array([0.17224810, 0.09170660, 0.06416938])
        RGB_c = colour_correction_Cheung2004(
            RGB, MATRIX_TEST, MATRIX_REFERENCE
        )

        RGB = np.tile(RGB, (6, 1))
        RGB_c = np.tile(RGB_c, (6, 1))
        np.testing.assert_almost_equal(
            colour_correction_Cheung2004(RGB, MATRIX_TEST, MATRIX_REFERENCE),
            RGB_c,
            decimal=7,
        )

        RGB = np.reshape(RGB, (2, 3, 3))
        RGB_c = np.reshape(RGB_c, (2, 3, 3))
        np.testing.assert_almost_equal(
            colour_correction_Cheung2004(RGB, MATRIX_TEST, MATRIX_REFERENCE),
            RGB_c,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_colour_correction_Cheung2004(self):  # pragma: no cover
        """
        Test :func:`colour.characterisation.correction.\
colour_correction_Cheung2004` definition nan support.
        """

        # NOTE: Hangs on "Linux".
        if platform.system() == "Linux":
            return

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = list(set(permutations(cases * 3, r=3)))[0:4]
        for case in cases:
            try:
                colour_correction_Cheung2004(
                    case,
                    np.vstack([case, case, case]),
                    np.transpose(np.vstack([case, case, case])),
                )
            except LinAlgError:
                pass


class TestColourCorrectionFinlayson2015(unittest.TestCase):
    """
    Define :func:`colour.characterisation.correction.\
colour_correction_Finlayson2015` definition unit tests methods.
    """

    def test_colour_correction_Finlayson2015(self):
        """
        Test :func:`colour.characterisation.correction.\
colour_correction_Finlayson2015` definition.
        """

        RGB = np.array([0.17224810, 0.09170660, 0.06416938])

        np.testing.assert_almost_equal(
            colour_correction_Finlayson2015(
                RGB, MATRIX_TEST, MATRIX_REFERENCE
            ),
            np.array([0.13348722, 0.08439216, 0.05990144]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            colour_correction_Finlayson2015(
                RGB, MATRIX_TEST, MATRIX_REFERENCE, degree=3
            ),
            np.array([0.13914542, 0.08602124, 0.06422973]),
            decimal=7,
        )

    def test_n_dimensional_colour_correction_Finlayson2015(self):
        """
        Test :func:`colour.characterisation.correction.\
colour_correction_Finlayson2015` definition n-dimensional support.
        """

        RGB = np.array([0.17224810, 0.09170660, 0.06416938])
        RGB_c = colour_correction_Finlayson2015(
            RGB, MATRIX_TEST, MATRIX_REFERENCE
        )

        RGB = np.tile(RGB, (6, 1))
        RGB_c = np.tile(RGB_c, (6, 1))
        np.testing.assert_almost_equal(
            colour_correction_Finlayson2015(
                RGB, MATRIX_TEST, MATRIX_REFERENCE
            ),
            RGB_c,
            decimal=7,
        )

        RGB = np.reshape(RGB, (2, 3, 3))
        RGB_c = np.reshape(RGB_c, (2, 3, 3))
        np.testing.assert_almost_equal(
            colour_correction_Finlayson2015(
                RGB, MATRIX_TEST, MATRIX_REFERENCE
            ),
            RGB_c,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_colour_correction_Finlayson2015(self):  # pragma: no cover
        """
                Test :func:`colour.characterisation.correction.
        colour_correction_Finlayson2015` definition nan support.
        """

        # NOTE: Hangs on "Linux".
        if platform.system() == "Linux":
            return

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = list(set(permutations(cases * 3, r=3)))[0:4]
        for case in cases:
            try:
                colour_correction_Finlayson2015(
                    case,
                    np.vstack([case, case, case]),
                    np.transpose(np.vstack([case, case, case])),
                )
            except LinAlgError:
                pass


class TestColourCorrectionVandermonde(unittest.TestCase):
    """
    Define :func:`colour.characterisation.correction.\
colour_correction_Vandermonde` definition unit tests methods.
    """

    def test_colour_correction_Vandermonde(self):
        """
        Test :func:`colour.characterisation.correction.\
colour_correction_Vandermonde` definition.
        """

        RGB = np.array([0.17224810, 0.09170660, 0.06416938])

        np.testing.assert_almost_equal(
            colour_correction_Vandermonde(RGB, MATRIX_TEST, MATRIX_REFERENCE),
            np.array([0.15034881, 0.10503956, 0.10512517]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            colour_correction_Vandermonde(
                RGB, MATRIX_TEST, MATRIX_REFERENCE, degree=3
            ),
            np.array([0.15747814, 0.10035799, 0.06616709]),
            decimal=7,
        )

    def test_n_dimensional_colour_correction_Vandermonde(self):
        """
        Test :func:`colour.characterisation.correction.\
colour_correction_Vandermonde` definition n-dimensional support.
        """

        RGB = np.array([0.17224810, 0.09170660, 0.06416938])
        RGB_c = colour_correction_Vandermonde(
            RGB, MATRIX_TEST, MATRIX_REFERENCE
        )

        RGB = np.tile(RGB, (6, 1))
        RGB_c = np.tile(RGB_c, (6, 1))
        np.testing.assert_almost_equal(
            colour_correction_Vandermonde(RGB, MATRIX_TEST, MATRIX_REFERENCE),
            RGB_c,
            decimal=7,
        )

        RGB = np.reshape(RGB, (2, 3, 3))
        RGB_c = np.reshape(RGB_c, (2, 3, 3))
        np.testing.assert_almost_equal(
            colour_correction_Vandermonde(RGB, MATRIX_TEST, MATRIX_REFERENCE),
            RGB_c,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_colour_correction_Vandermonde(self):  # pragma: no cover
        """
        Test :func:`colour.characterisation.correction.\
colour_correction_Vandermonde` definition nan support.
        """

        # NOTE: Hangs on "Linux".
        if platform.system() == "Linux":
            return

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = list(set(permutations(cases * 3, r=3)))[0:4]
        for case in cases:
            try:
                colour_correction_Vandermonde(
                    case,
                    np.vstack([case, case, case]),
                    np.transpose(np.vstack([case, case, case])),
                )
            except LinAlgError:
                pass


if __name__ == "__main__":
    unittest.main()



"""
RGB Colourspace and Transformations
===================================

Defines the :class:`colour.RGB_Colourspace` class for the *RGB* colourspaces
datasets from :mod:`colour.models.datasets.aces_rgb`, etc... and the following
*RGB* colourspace transformations or helper definitions:

-   :func:`colour.XYZ_to_RGB`
-   :func:`colour.RGB_to_XYZ`
-   :func:`colour.matrix_RGB_to_RGB`
-   :func:`colour.RGB_to_RGB`

References
----------
-   :cite:`InternationalElectrotechnicalCommission1999a` : International
    Electrotechnical Commission. (1999). IEC 61966-2-1:1999 - Multimedia
    systems and equipment - Colour measurement and management - Part 2-1:
    Colour management - Default RGB colour space - sRGB (p. 51).
    https://webstore.iec.ch/publication/6169
-   :cite:`Panasonic2014a` : Panasonic. (2014). VARICAM V-Log/V-Gamut (pp.
    1-7).
    http://pro-av.panasonic.net/en/varicam/common/pdf/VARICAM_V-Log_V-Gamut.pdf
"""

from __future__ import annotations

import numpy as np
from copy import deepcopy

from colour.adaptation import matrix_chromatic_adaptation_VonKries
from colour.algebra import matrix_dot, vector_dot
from colour.hints import (
    Any,
    ArrayLike,
    Boolean,
    Callable,
    Literal,
    NDArray,
    Optional,
    Union,
    cast,
)
from colour.models import xy_to_XYZ, xy_to_xyY, xyY_to_XYZ
from colour.models.rgb import (
    chromatically_adapted_primaries,
    normalised_primary_matrix,
)
from colour.utilities import (
    as_float_array,
    attest,
    domain_range_scale,
    filter_kwargs,
    from_range_1,
    optional,
    to_domain_1,
    is_string,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "RGB_Colourspace",
    "XYZ_to_RGB",
    "RGB_to_XYZ",
    "matrix_RGB_to_RGB",
    "RGB_to_RGB",
]


class RGB_Colourspace:
    """
    Implement support for the *RGB* colourspaces datasets from
    :mod:`colour.models.datasets.aces_rgb`, etc....

    Colour science literature related to *RGB* colourspaces and encodings
    defines their dataset using different degree of precision or rounding.
    While instances where a whitepoint is being defined with a value
    different than its canonical agreed one are rare, it is however very
    common to have normalised primary matrices rounded at different
    decimals. This can yield large discrepancies in computations.

    Such an occurrence is the *V-Gamut* colourspace white paper, that defines
    the *V-Gamut* to *ITU-R BT.709* conversion matrix as follows::

        [[ 1.806576 -0.695697 -0.110879]
         [-0.170090  1.305955 -0.135865]
         [-0.025206 -0.154468  1.179674]]

    Computing this matrix using *ITU-R BT.709* colourspace derived normalised
    primary matrix yields::

        [[ 1.8065736 -0.6956981 -0.1108786]
         [-0.1700890  1.3059548 -0.1358648]
         [-0.0252057 -0.1544678  1.1796737]]

    The latter matrix is almost equals with the former, however performing the
    same computation using *IEC 61966-2-1:1999* *sRGB* colourspace normalised
    primary matrix introduces severe disparities::

        [[ 1.8063853 -0.6956147 -0.1109453]
         [-0.1699311  1.3058387 -0.1358616]
         [-0.0251630 -0.1544899  1.1797117]]

    In order to provide support for both literature defined dataset and
    accurate computations enabling transformations without loss of precision,
    the :class:`colour.RGB_Colourspace` class provides two sets of
    transformation matrices:

        -   Instantiation transformation matrices
        -   Derived transformation matrices

    Upon instantiation, the :class:`colour.RGB_Colourspace` class stores the
    given ``matrix_RGB_to_XYZ`` and ``matrix_XYZ_to_RGB`` arguments and also
    computes their derived counterpart using the ``primaries`` and
    ``whitepoint`` arguments.

    Whether the initialisation or derived matrices are used in subsequent
    computations is dependent on the
    :attr:`colour.RGB_Colourspace.use_derived_matrix_RGB_to_XYZ` and
    :attr:`colour.RGB_Colourspace.use_derived_matrix_XYZ_to_RGB` attribute
    values.

    Parameters
    ----------
    name
        *RGB* colourspace name.
    primaries
        *RGB* colourspace primaries.
    whitepoint
        *RGB* colourspace whitepoint.
    whitepoint_name
        *RGB* colourspace whitepoint name.
    matrix_RGB_to_XYZ
        Transformation matrix from colourspace to *CIE XYZ* tristimulus values.
    matrix_XYZ_to_RGB
        Transformation matrix from *CIE XYZ* tristimulus values to colourspace.
    cctf_encoding
        Encoding colour component transfer function (Encoding CCTF) /
        opto-electronic transfer function (OETF) that maps estimated
        tristimulus values in a scene to :math:`R'G'B'` video component signal
        value.
    cctf_decoding
        Decoding colour component transfer function (Decoding CCTF) /
        electro-optical transfer function (EOTF) that maps an
        :math:`R'G'B'` video component signal value to tristimulus values at
        the display.
    use_derived_matrix_RGB_to_XYZ
        Whether to use the instantiation time normalised primary matrix or to
        use a computed derived normalised primary matrix.
    use_derived_matrix_XYZ_to_RGB
        Whether to use the instantiation time inverse normalised primary
        matrix or to use a computed derived inverse normalised primary matrix.

    Attributes
    ----------
    -   :attr:`~colour.RGB_Colourspace.name`
    -   :attr:`~colour.RGB_Colourspace.primaries`
    -   :attr:`~colour.RGB_Colourspace.whitepoint`
    -   :attr:`~colour.RGB_Colourspace.whitepoint_name`
    -   :attr:`~colour.RGB_Colourspace.matrix_RGB_to_XYZ`
    -   :attr:`~colour.RGB_Colourspace.matrix_XYZ_to_RGB`
    -   :attr:`~colour.RGB_Colourspace.cctf_encoding`
    -   :attr:`~colour.RGB_Colourspace.cctf_decoding`
    -   :attr:`~colour.RGB_Colourspace.use_derived_matrix_RGB_to_XYZ`
    -   :attr:`~colour.RGB_Colourspace.use_derived_matrix_XYZ_to_RGB`

    Methods
    -------
    -   :attr:`~colour.RGB_Colourspace.__init__`
    -   :attr:`~colour.RGB_Colourspace.__str__`
    -   :attr:`~colour.RGB_Colourspace.__repr__`
    -   :attr:`~colour.RGB_Colourspace.use_derived_transformation_matrices`
    -   :attr:`~colour.RGB_Colourspace.chromatically_adapt`
    -   :attr:`~colour.RGB_Colourspace.copy`

    Notes
    -----
    -   The normalised primary matrix defined by
        :attr:`colour.RGB_Colourspace.matrix_RGB_to_XYZ` property is treated
        as the prime matrix from which the inverse will be calculated as
        required by the internal derivation mechanism. This behaviour has been
        chosen in accordance with literature where commonly a *RGB* colourspace
        is defined by its normalised primary matrix as it is directly computed
        from the chosen primaries and whitepoint.

    References
    ----------
    :cite:`InternationalElectrotechnicalCommission1999a`,
    :cite:`Panasonic2014a`

    Examples
    --------
    >>> p = np.array([0.73470, 0.26530, 0.00000, 1.00000, 0.00010, -0.07700])
    >>> whitepoint = np.array([0.32168, 0.33767])
    >>> matrix_RGB_to_XYZ = np.identity(3)
    >>> matrix_XYZ_to_RGB = np.identity(3)
    >>> colourspace = RGB_Colourspace('RGB Colourspace', p, whitepoint, 'ACES',
    ...                               matrix_RGB_to_XYZ, matrix_XYZ_to_RGB)
    >>> colourspace.matrix_RGB_to_XYZ
    array([[ 1.,  0.,  0.],
           [ 0.,  1.,  0.],
           [ 0.,  0.,  1.]])
    >>> colourspace.matrix_XYZ_to_RGB
    array([[ 1.,  0.,  0.],
           [ 0.,  1.,  0.],
           [ 0.,  0.,  1.]])
    >>> colourspace.use_derived_transformation_matrices(True)
    >>> colourspace.matrix_RGB_to_XYZ  # doctest: +ELLIPSIS
    array([[  9.5255239...e-01,   0.0000000...e+00,   9.3678631...e-05],
           [  3.4396645...e-01,   7.2816609...e-01,  -7.2132546...e-02],
           [  0.0000000...e+00,   0.0000000...e+00,   1.0088251...e+00]])
    >>> colourspace.matrix_XYZ_to_RGB  # doctest: +ELLIPSIS
    array([[  1.0498110...e+00,   0.0000000...e+00,  -9.7484540...e-05],
           [ -4.9590302...e-01,   1.3733130...e+00,   9.8240036...e-02],
           [  0.0000000...e+00,   0.0000000...e+00,   9.9125201...e-01]])
    >>> colourspace.use_derived_matrix_RGB_to_XYZ = False
    >>> colourspace.matrix_RGB_to_XYZ
    array([[ 1.,  0.,  0.],
           [ 0.,  1.,  0.],
           [ 0.,  0.,  1.]])
    >>> colourspace.use_derived_matrix_XYZ_to_RGB = False
    >>> colourspace.matrix_XYZ_to_RGB
    array([[ 1.,  0.,  0.],
           [ 0.,  1.,  0.],
           [ 0.,  0.,  1.]])
    """

    def __init__(
        self,
        name: str,
        primaries: ArrayLike,
        whitepoint: ArrayLike,
        whitepoint_name: Optional[str] = None,
        matrix_RGB_to_XYZ: Optional[ArrayLike] = None,
        matrix_XYZ_to_RGB: Optional[ArrayLike] = None,
        cctf_encoding: Optional[Callable] = None,
        cctf_decoding: Optional[Callable] = None,
        use_derived_matrix_RGB_to_XYZ: Boolean = False,
        use_derived_matrix_XYZ_to_RGB: Boolean = False,
    ):
        self._derived_matrix_RGB_to_XYZ: NDArray = np.array([])
        self._derived_matrix_XYZ_to_RGB: NDArray = np.array([])

        self._name: str = f"{self.__class__.__name__} ({id(self)})"
        self.name = name
        self._primaries: NDArray = np.array([])
        self.primaries = primaries  # type: ignore[assignment]
        self._whitepoint: NDArray = np.array([])
        self.whitepoint = whitepoint  # type: ignore[assignment]
        self._whitepoint_name: Optional[str] = None
        self.whitepoint_name = whitepoint_name
        self._matrix_RGB_to_XYZ: Optional[NDArray] = None
        self.matrix_RGB_to_XYZ = matrix_RGB_to_XYZ  # type: ignore[assignment]
        self._matrix_XYZ_to_RGB: Optional[NDArray] = None
        self.matrix_XYZ_to_RGB = matrix_XYZ_to_RGB  # type: ignore[assignment]
        self._cctf_encoding: Optional[Callable] = None
        self.cctf_encoding = cctf_encoding
        self._cctf_decoding: Optional[Callable] = None
        self.cctf_decoding = cctf_decoding
        self._use_derived_matrix_RGB_to_XYZ: Boolean = False
        self.use_derived_matrix_RGB_to_XYZ = use_derived_matrix_RGB_to_XYZ
        self._use_derived_matrix_XYZ_to_RGB: Boolean = False
        self.use_derived_matrix_XYZ_to_RGB = use_derived_matrix_XYZ_to_RGB

    @property
    def name(self) -> str:
        """
        Getter and setter property for the name.

        Parameters
        ----------
        value
            Value to set the name with.

        Returns
        -------
        :class:`str`
            *RGB* colourspace name.
        """

        return self._name

    @name.setter
    def name(self, value: str):
        """Setter for the **self.name** property."""

        attest(
            is_string(value),
            f'"name" property: "{value}" type is not "str"!',
        )

        self._name = value

    @property
    def primaries(self) -> NDArray:
        """
        Getter and setter property for the primaries.

        Parameters
        ----------
        value
            Value to set the primaries with.

        Returns
        -------
        :class:`numpy.ndarray`
            *RGB* colourspace primaries.
        """

        return self._primaries

    @primaries.setter
    def primaries(self, value: ArrayLike):
        """Setter for the **self.primaries** property."""

        attest(
            isinstance(value, (tuple, list, np.ndarray, np.matrix)),
            f'"matrix_XYZ_to_RGB" property: "{value!r}" is not a "tuple", '
            f'"list", "ndarray" or "matrix" instance!',
        )

        value = as_float_array(value)

        value = np.reshape(value, (3, 2))

        self._primaries = value

        self._derive_transformation_matrices()

    @property
    def whitepoint(self) -> NDArray:
        """
        Getter and setter property for the whitepoint.

        Parameters
        ----------
        value
            Value to set the whitepoint with.

        Returns
        -------
        :class:`numpy.ndarray`
            *RGB* colourspace whitepoint.
        """

        return self._whitepoint

    @whitepoint.setter
    def whitepoint(self, value: ArrayLike):
        """Setter for the **self.whitepoint** property."""

        attest(
            isinstance(value, (tuple, list, np.ndarray, np.matrix)),
            f'"matrix_XYZ_to_RGB" property: "{value!r}" is not a "tuple", '
            f'"list", "ndarray" or "matrix" instance!',
        )

        value = as_float_array(value)

        self._whitepoint = value

        self._derive_transformation_matrices()

    @property
    def whitepoint_name(self) -> Optional[str]:
        """
        Getter and setter property for the whitepoint_name.

        Parameters
        ----------
        value
            Value to set the whitepoint_name with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            *RGB* colourspace whitepoint name.
        """

        return self._whitepoint_name

    @whitepoint_name.setter
    def whitepoint_name(self, value: Optional[str]):
        """Setter for the **self.whitepoint_name** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"whitepoint_name" property: "{value}" type is not "str"!',
            )

        self._whitepoint_name = value

    @property
    def matrix_RGB_to_XYZ(self) -> NDArray:
        """
        Getter and setter property for the transformation matrix from
        colourspace to *CIE XYZ* tristimulus values.

        Parameters
        ----------
        value
            Transformation matrix from colourspace to *CIE XYZ* tristimulus
            values.

        Returns
        -------
        :class:`numpy.ndarray`
            Transformation matrix from colourspace to *CIE XYZ* tristimulus
            values.
        """

        if (
            self._matrix_RGB_to_XYZ is None
            or self._use_derived_matrix_RGB_to_XYZ
        ):
            return self._derived_matrix_RGB_to_XYZ
        else:
            return self._matrix_RGB_to_XYZ

    @matrix_RGB_to_XYZ.setter
    def matrix_RGB_to_XYZ(self, value: Optional[ArrayLike]):
        """Setter for the **self.matrix_RGB_to_XYZ** property."""

        if value is not None:
            attest(
                isinstance(value, (tuple, list, np.ndarray, np.matrix)),
                f'"matrix_RGB_to_XYZ" property: "{value!r}" is not a "tuple", '
                f'"list", "ndarray" or "matrix" instance!',
            )

            value = as_float_array(value)

        self._matrix_RGB_to_XYZ = value

    @property
    def matrix_XYZ_to_RGB(self) -> NDArray:
        """
        Getter and setter property for the transformation matrix from *CIE XYZ*
        tristimulus values to colourspace.

        Parameters
        ----------
        value
            Transformation matrix from *CIE XYZ* tristimulus values to
            colourspace.

        Returns
        -------
        :class:`numpy.ndarray`
            Transformation matrix from *CIE XYZ* tristimulus values to
            colourspace.
        """

        if (
            self._matrix_XYZ_to_RGB is None
            or self._use_derived_matrix_XYZ_to_RGB
        ):
            return self._derived_matrix_XYZ_to_RGB
        else:
            return self._matrix_XYZ_to_RGB

    @matrix_XYZ_to_RGB.setter
    def matrix_XYZ_to_RGB(self, value: Optional[ArrayLike]):
        """Setter for the **self.matrix_XYZ_to_RGB** property."""

        if value is not None:
            attest(
                isinstance(value, (tuple, list, np.ndarray, np.matrix)),
                f'"matrix_XYZ_to_RGB" property: "{value!r}" is not a "tuple", '
                f'"list", "ndarray" or "matrix" instance!',
            )

            value = as_float_array(value)

        self._matrix_XYZ_to_RGB = value

    @property
    def cctf_encoding(self) -> Optional[Callable]:
        """
        Getter and setter property for the encoding colour component transfer
        function (Encoding CCTF) / opto-electronic transfer function
        (OETF).

        Parameters
        ----------
        value
            Encoding colour component transfer function (Encoding CCTF) /
            opto-electronic transfer function (OETF).

        Returns
        -------
        :py:data:`None` or Callable
            Encoding colour component transfer function (Encoding CCTF) /
            opto-electronic transfer function (OETF).
        """

        return self._cctf_encoding

    @cctf_encoding.setter
    def cctf_encoding(self, value: Optional[Callable]):
        """Setter for the **self.cctf_encoding** property."""

        if value is not None:
            attest(
                hasattr(value, "__call__"),
                f'"cctf_encoding" property: "{value}" is not callable!',
            )

        self._cctf_encoding = value

    @property
    def cctf_decoding(self) -> Optional[Callable]:
        """
        Getter and setter property for the decoding colour component transfer
        function (Decoding CCTF) / electro-optical transfer function
        (EOTF).

        Parameters
        ----------
        value
            Decoding colour component transfer function (Decoding CCTF) /
            electro-optical transfer function (EOTF).

        Returns
        -------
        :py:data:`None` or Callable
            Decoding colour component transfer function (Decoding CCTF) /
            electro-optical transfer function (EOTF).
        """

        return self._cctf_decoding

    @cctf_decoding.setter
    def cctf_decoding(self, value: Optional[Callable]):
        """Setter for the **self.cctf_decoding** property."""

        if value is not None:
            attest(
                hasattr(value, "__call__"),
                f'"cctf_decoding" property: "{value}" is not callable!',
            )

        self._cctf_decoding = value

    @property
    def use_derived_matrix_RGB_to_XYZ(self) -> Boolean:
        """
        Getter and setter property for whether to use the instantiation time
        normalised primary matrix or to use a computed derived normalised
        primary matrix.

        Parameters
        ----------
        value
            Whether to use the instantiation time normalised primary matrix or
            to use a computed derived normalised primary matrix.

        Returns
        -------
        :class:`bool`
            Whether to use the instantiation time normalised primary matrix or
            to use a computed derived normalised primary matrix.
        """

        return self._use_derived_matrix_RGB_to_XYZ

    @use_derived_matrix_RGB_to_XYZ.setter
    def use_derived_matrix_RGB_to_XYZ(self, value: Boolean):
        """Setter for the **self.use_derived_matrix_RGB_to_XYZ** property."""

        attest(
            isinstance(value, (bool, np.bool_)),
            f'"use_derived_matrix_RGB_to_XYZ" property: "{value}" is not a '
            f'"bool"!',
        )

        self._use_derived_matrix_RGB_to_XYZ = value

    @property
    def use_derived_matrix_XYZ_to_RGB(self) -> Boolean:
        """
        Getter and setter property for Whether to use the instantiation time
        inverse normalised primary matrix or to use a computed derived inverse
        normalised primary matrix.

        Parameters
        ----------
        value
            Whether to use the instantiation time inverse normalised primary
            matrix or to use a computed derived inverse normalised primary
            matrix.

        Returns
        -------
        :class:`bool`
            Whether to use the instantiation time inverse normalised primary
            matrix or to use a computed derived inverse normalised primary
            matrix.
        """

        return self._use_derived_matrix_XYZ_to_RGB

    @use_derived_matrix_XYZ_to_RGB.setter
    def use_derived_matrix_XYZ_to_RGB(self, value: Boolean):
        """Setter for the **self.use_derived_matrix_XYZ_to_RGB** property."""

        attest(
            isinstance(value, (bool, np.bool_)),
            f'"use_derived_matrix_XYZ_to_RGB" property: "{value}" is not a '
            f'"bool"!',
        )

        self._use_derived_matrix_XYZ_to_RGB = value

    def __str__(self) -> str:
        """
        Return a formatted string representation of the *RGB* colourspace.

        Returns
        -------
        :class:`str`
            Formatted string representation.

        Examples
        --------
        >>> p = np.array(
        ...     [0.73470, 0.26530, 0.00000, 1.00000, 0.00010, -0.07700])
        >>> whitepoint = np.array([0.32168, 0.33767])
        >>> matrix_RGB_to_XYZ = np.identity(3)
        >>> matrix_XYZ_to_RGB = np.identity(3)
        >>> cctf_encoding = lambda x: x
        >>> cctf_decoding = lambda x: x
        >>> print(RGB_Colourspace('RGB Colourspace', p, whitepoint, 'ACES',
        ...                       matrix_RGB_to_XYZ, matrix_XYZ_to_RGB,
        ...                       cctf_encoding, cctf_decoding))
        ... # doctest: +ELLIPSIS
        RGB Colourspace
        ---------------
        <BLANKLINE>
        Primaries          : [[  7.34700000e-01   2.65300000e-01]
                              [  0.00000000e+00   1.00000000e+00]
                              [  1.00000000e-04  -7.70000000e-02]]
        Whitepoint         : [ 0.32168  0.33767]
        Whitepoint Name    : ACES
        Encoding CCTF      : <function <lambda> at 0x...>
        Decoding CCTF      : <function <lambda> at 0x...>
        NPM                : [[ 1.  0.  0.]
                              [ 0.  1.  0.]
                              [ 0.  0.  1.]]
        NPM -1             : [[ 1.  0.  0.]
                              [ 0.  1.  0.]
                              [ 0.  0.  1.]]
        Derived NPM        : \
[[  9.5255239...e-01   0.0000000...e+00   9.3678631...e-05]
                             \
 [  3.4396645...e-01   7.2816609...e-01  -7.2132546...e-02]
                             \
 [  0.0000000...e+00   0.0000000...e+00   1.0088251...e+00]]
        Derived NPM -1     : \
[[  1.0498110...e+00   0.0000000...e+00  -9.7484540...e-05]
                             \
 [ -4.9590302...e-01   1.3733130...e+00   9.8240036...e-02]
                             \
 [  0.0000000...e+00   0.0000000...e+00   9.9125201...e-01]]
        Use Derived NPM    : False
        Use Derived NPM -1 : False
        """

        def _indent_array(a: Optional[NDArray]) -> str:
            """Indent given array string representation."""

            return str(a).replace(" [", " " * 22 + "[")

        return (
            f"{self.name}\n"
            f'{"-" * len(self.name)}\n\n'
            f"Primaries          : {_indent_array(self.primaries)}\n"
            f"Whitepoint         : {self.whitepoint}\n"
            f"Whitepoint Name    : {self.whitepoint_name}\n"
            f"Encoding CCTF      : {self.cctf_encoding}\n"
            f"Decoding CCTF      : {self.cctf_decoding}\n"
            f"NPM                : {_indent_array(self._matrix_RGB_to_XYZ)}\n"
            f"NPM -1             : {_indent_array(self._matrix_XYZ_to_RGB)}\n"
            f"Derived NPM        : {_indent_array(self._derived_matrix_RGB_to_XYZ)}\n"
            f"Derived NPM -1     : {_indent_array(self._derived_matrix_XYZ_to_RGB)}\n"
            f"Use Derived NPM    : {self.use_derived_matrix_RGB_to_XYZ}\n"
            f"Use Derived NPM -1 : {self.use_derived_matrix_XYZ_to_RGB}"
        )

    def __repr__(self) -> str:
        """
        Return an (almost) evaluable string representation of the *RGB*
        colourspace.

        Returns
        -------
        :class`str`
            (Almost) evaluable string representation.

        Examples
        --------
        >>> p = np.array(
        ...     [0.73470, 0.26530, 0.00000, 1.00000, 0.00010, -0.07700])
        >>> whitepoint = np.array([0.32168, 0.33767])
        >>> matrix_RGB_to_XYZ = np.identity(3)
        >>> matrix_XYZ_to_RGB = np.identity(3)
        >>> cctf_encoding = lambda x: x
        >>> cctf_decoding = lambda x: x
        >>> RGB_Colourspace('RGB Colourspace', p, whitepoint, 'ACES',
        ...                 matrix_RGB_to_XYZ, matrix_XYZ_to_RGB,
        ...                 cctf_encoding, cctf_decoding)
        ... # doctest: +ELLIPSIS
        RGB_Colourspace(RGB Colourspace,
                        [[  7.34700000e-01,   2.65300000e-01],
                         [  0.00000000e+00,   1.00000000e+00],
                         [  1.00000000e-04,  -7.70000000e-02]],
                        [ 0.32168,  0.33767],
                        ACES,
                        [[ 1.,  0.,  0.],
                         [ 0.,  1.,  0.],
                         [ 0.,  0.,  1.]],
                        [[ 1.,  0.,  0.],
                         [ 0.,  1.,  0.],
                         [ 0.,  0.,  1.]],
                        <function <lambda> at 0x...>,
                        <function <lambda> at 0x...>,
                        False,
                        False)
        """

        def _indent_array(a: Optional[NDArray]) -> str:
            """Indent given array evaluable string representation."""

            representation = repr(a).replace(" [", f"{' ' * 11}[")
            representation = representation.replace("array(", " " * 16)
            return representation.replace(")", "")

        indentation = " " * 16
        return (
            f"RGB_Colourspace({self.name},\n"
            f"{_indent_array(self.primaries)},\n"
            f"{_indent_array(self.whitepoint)},\n"
            f"{indentation}{self.whitepoint_name},\n"
            f"{_indent_array(self.matrix_RGB_to_XYZ)},\n"
            f"{_indent_array(self.matrix_XYZ_to_RGB)},\n"
            f"{indentation}{self.cctf_encoding},\n"
            f"{indentation}{self.cctf_decoding},\n"
            f"{indentation}{self.use_derived_matrix_RGB_to_XYZ},\n"
            f"{indentation}{self.use_derived_matrix_XYZ_to_RGB})"
        )

    def _derive_transformation_matrices(self):
        """
        Compute the derived transformations matrices, the normalised primary
        matrix and its inverse.
        """

        if hasattr(self, "_primaries") and hasattr(self, "_whitepoint"):
            if self._primaries is not None and self._whitepoint is not None:
                npm = normalised_primary_matrix(
                    self._primaries, self._whitepoint
                )

                self._derived_matrix_RGB_to_XYZ = npm
                self._derived_matrix_XYZ_to_RGB = np.linalg.inv(npm)

    def use_derived_transformation_matrices(self, usage: Boolean = True):
        """
        Enable or disables usage of both derived transformations matrices,
        the normalised primary matrix and its inverse in subsequent
        computations.

        Parameters
        ----------
        usage
            Whether to use the derived transformations matrices.
        """

        self.use_derived_matrix_RGB_to_XYZ = usage
        self.use_derived_matrix_XYZ_to_RGB = usage

    def chromatically_adapt(
        self,
        whitepoint: ArrayLike,
        whitepoint_name: Optional[str] = None,
        chromatic_adaptation_transform: Union[
            Literal[
                "Bianco 2010",
                "Bianco PC 2010",
                "Bradford",
                "CAT02 Brill 2008",
                "CAT02",
                "CAT16",
                "CMCCAT2000",
                "CMCCAT97",
                "Fairchild",
                "Sharp",
                "Von Kries",
                "XYZ Scaling",
            ],
            str,
        ] = "CAT02",
    ) -> RGB_Colourspace:
        """
        Chromatically adapt the *RGB* colourspace *primaries* :math:`xy`
        chromaticity coordinates from *RGB* colourspace whitepoint to reference
        ``whitepoint``.

        Parameters
        ----------
        whitepoint
            Reference illuminant / whitepoint :math:`xy` chromaticity
            coordinates.
        whitepoint_name
            Reference illuminant / whitepoint name.
        chromatic_adaptation_transform
            *Chromatic adaptation* transform.

        Returns
        -------
        :class:`colour.RGB_Colourspace`
            Chromatically adapted *RGB* colourspace.

        Examples
        --------
        >>> p = np.array(
        ...     [0.73470, 0.26530, 0.00000, 1.00000, 0.00010, -0.07700])
        >>> w_t = np.array([0.32168, 0.33767])
        >>> w_r = np.array([0.31270, 0.32900])
        >>> colourspace = RGB_Colourspace('RGB Colourspace', p, w_t, 'D65')
        >>> print(colourspace.chromatically_adapt(w_r, 'D50', 'Bradford'))
        ... # doctest: +ELLIPSIS
        RGB Colourspace - Chromatically Adapted to 'D50'
        ------------------------------------------------
        <BLANKLINE>
        Primaries          : [[ 0.73485524  0.26422533]
                              [-0.00617091  1.01131496]
                              [ 0.01596756 -0.0642355 ]]
        Whitepoint         : [ 0.3127  0.329 ]
        Whitepoint Name    : D50
        Encoding CCTF      : None
        Decoding CCTF      : None
        NPM                : None
        NPM -1             : None
        Derived NPM        : [[ 0.93827985 -0.00445145  0.01662752]
                              [ 0.33736889  0.72952157 -0.06689046]
                              [ 0.00117395 -0.00371071  1.09159451]]
        Derived NPM -1     : [[ 1.06349549  0.00640891 -0.01580679]
                              [-0.49207413  1.36822341  0.09133709]
                              [-0.00281646  0.00464417  0.91641857]]
        Use Derived NPM    : True
        Use Derived NPM -1 : True
        """

        colourspace = self.copy()

        colourspace.primaries = chromatically_adapted_primaries(
            colourspace.primaries,
            colourspace.whitepoint,
            whitepoint,
            chromatic_adaptation_transform,
        )
        colourspace.whitepoint = whitepoint  # type: ignore[assignment]
        colourspace.whitepoint_name = whitepoint_name

        colourspace._matrix_RGB_to_XYZ = None
        colourspace._matrix_XYZ_to_RGB = None
        colourspace._derive_transformation_matrices()
        colourspace.use_derived_transformation_matrices()

        colourspace.name = (
            f"{colourspace.name} - Chromatically Adapted to "
            f"{cast(str, optional(whitepoint_name, whitepoint))!r}"
        )

        return colourspace

    def copy(self) -> RGB_Colourspace:
        """
        Return a copy of the *RGB* colourspace.

        Returns
        -------
        :class:`colour.RGB_Colourspace`
            *RGB* colourspace copy.
        """

        return deepcopy(self)


def XYZ_to_RGB(
    XYZ: ArrayLike,
    illuminant_XYZ: ArrayLike,
    illuminant_RGB: ArrayLike,
    matrix_XYZ_to_RGB: ArrayLike,
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
    cctf_encoding: Optional[Callable] = None,
) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to *RGB* colourspace array.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.
    illuminant_XYZ
        *CIE xy* chromaticity coordinates or *CIE xyY* colourspace array of the
        *illuminant* for the input *CIE XYZ* tristimulus values.
    illuminant_RGB
        *CIE xy* chromaticity coordinates or *CIE xyY* colourspace array of the
        *illuminant* for the output *RGB* colourspace array.
    matrix_XYZ_to_RGB
        Matrix converting the *CIE XYZ* tristimulus values to *RGB* colourspace
        array, i.e. the inverse *Normalised Primary Matrix* (NPM).
    chromatic_adaptation_transform
        *Chromatic adaptation* transform, if *None* no chromatic adaptation is
        performed.
    cctf_encoding
        Encoding colour component transfer function (Encoding CCTF) or
        opto-electronic transfer function (OETF).

    Returns
    -------
    :class:`numpy.ndarray`
        *RGB* colourspace array.

    Notes
    -----
    +--------------------+-----------------------+---------------+
    | **Domain**         | **Scale - Reference** | **Scale - 1** |
    +====================+=======================+===============+
    | ``XYZ``            | [0, 1]                | [0, 1]        |
    +--------------------+-----------------------+---------------+
    | ``illuminant_XYZ`` | [0, 1]                | [0, 1]        |
    +--------------------+-----------------------+---------------+
    | ``illuminant_RGB`` | [0, 1]                | [0, 1]        |
    +--------------------+-----------------------+---------------+

    +--------------------+-----------------------+---------------+
    | **Range**          | **Scale - Reference** | **Scale - 1** |
    +====================+=======================+===============+
    | ``RGB``            | [0, 1]                | [0, 1]        |
    +--------------------+-----------------------+---------------+

    Examples
    --------
    >>> XYZ = np.array([0.21638819, 0.12570000, 0.03847493])
    >>> illuminant_XYZ = np.array([0.34570, 0.35850])
    >>> illuminant_RGB = np.array([0.31270, 0.32900])
    >>> chromatic_adaptation_transform = 'Bradford'
    >>> matrix_XYZ_to_RGB = np.array(
    ...     [[3.24062548, -1.53720797, -0.49862860],
    ...      [-0.96893071, 1.87575606, 0.04151752],
    ...      [0.05571012, -0.20402105, 1.05699594]]
    ... )
    >>> XYZ_to_RGB(XYZ, illuminant_XYZ, illuminant_RGB, matrix_XYZ_to_RGB,
    ...            chromatic_adaptation_transform)  # doctest: +ELLIPSIS
    array([ 0.4559557...,  0.0303970...,  0.0408724...])
    """

    XYZ = to_domain_1(XYZ)

    if chromatic_adaptation_transform is not None:
        M_CAT = matrix_chromatic_adaptation_VonKries(
            xyY_to_XYZ(xy_to_xyY(illuminant_XYZ)),
            xyY_to_XYZ(xy_to_xyY(illuminant_RGB)),
            transform=chromatic_adaptation_transform,
        )

        XYZ = vector_dot(M_CAT, XYZ)

    RGB = vector_dot(matrix_XYZ_to_RGB, XYZ)

    if cctf_encoding is not None:
        with domain_range_scale("ignore"):
            RGB = cctf_encoding(RGB)

    return from_range_1(RGB)


def RGB_to_XYZ(
    RGB: ArrayLike,
    illuminant_RGB: ArrayLike,
    illuminant_XYZ: ArrayLike,
    matrix_RGB_to_XYZ: ArrayLike,
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
    cctf_decoding: Optional[Callable] = None,
) -> NDArray:
    """
    Convert given *RGB* colourspace array to *CIE XYZ* tristimulus values.

    Parameters
    ----------
    RGB
        *RGB* colourspace array.
    illuminant_RGB
        *CIE xy* chromaticity coordinates or *CIE xyY* colourspace array of the
        *illuminant* for the input *RGB* colourspace array.
    illuminant_XYZ
        *CIE xy* chromaticity coordinates or *CIE xyY* colourspace array of the
        *illuminant* for the output *CIE XYZ* tristimulus values.
    matrix_RGB_to_XYZ
        Matrix converting the *RGB* colourspace array to *CIE XYZ* tristimulus
        values, i.e. the *Normalised Primary Matrix* (NPM).
    chromatic_adaptation_transform
        *Chromatic adaptation* transform, if *None* no chromatic adaptation is
        performed.
    cctf_decoding
        Decoding colour component transfer function (Decoding CCTF) or
        electro-optical transfer function (EOTF).

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +--------------------+-----------------------+---------------+
    | **Domain**         | **Scale - Reference** | **Scale - 1** |
    +====================+=======================+===============+
    | ``RGB``            | [0, 1]                | [0, 1]        |
    +--------------------+-----------------------+---------------+
    | ``illuminant_XYZ`` | [0, 1]                | [0, 1]        |
    +--------------------+-----------------------+---------------+
    | ``illuminant_RGB`` | [0, 1]                | [0, 1]        |
    +--------------------+-----------------------+---------------+

    +--------------------+-----------------------+---------------+
    | **Range**          | **Scale - Reference** | **Scale - 1** |
    +====================+=======================+===============+
    | ``XYZ``            | [0, 1]                | [0, 1]        |
    +--------------------+-----------------------+---------------+

    Examples
    --------
    >>> RGB = np.array([0.45595571, 0.03039702, 0.04087245])
    >>> illuminant_RGB = np.array([0.31270, 0.32900])
    >>> illuminant_XYZ = np.array([0.34570, 0.35850])
    >>> chromatic_adaptation_transform = 'Bradford'
    >>> matrix_RGB_to_XYZ = np.array(
    ...     [[0.41240000, 0.35760000, 0.18050000],
    ...      [0.21260000, 0.71520000, 0.07220000],
    ...      [0.01930000, 0.11920000, 0.95050000]]
    ... )
    >>> RGB_to_XYZ(RGB, illuminant_RGB, illuminant_XYZ, matrix_RGB_to_XYZ,
    ...            chromatic_adaptation_transform)  # doctest: +ELLIPSIS
    array([ 0.2163881...,  0.1257    ,  0.0384749...])
    """

    RGB = to_domain_1(RGB)

    if cctf_decoding is not None:
        with domain_range_scale("ignore"):
            RGB = cctf_decoding(RGB)

    XYZ = vector_dot(matrix_RGB_to_XYZ, RGB)

    if chromatic_adaptation_transform is not None:
        M_CAT = matrix_chromatic_adaptation_VonKries(
            xyY_to_XYZ(xy_to_xyY(illuminant_RGB)),
            xyY_to_XYZ(xy_to_xyY(illuminant_XYZ)),
            transform=chromatic_adaptation_transform,
        )

        XYZ = vector_dot(M_CAT, XYZ)

    return from_range_1(XYZ)


def matrix_RGB_to_RGB(
    input_colourspace: RGB_Colourspace,
    output_colourspace: RGB_Colourspace,
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
) -> NDArray:
    """
    Compute the matrix :math:`M` converting from given input *RGB*
    colourspace to output *RGB* colourspace using given *chromatic
    adaptation* method.

    Parameters
    ----------
    input_colourspace
        *RGB* input colourspace.
    output_colourspace
        *RGB* output colourspace.
    chromatic_adaptation_transform
        *Chromatic adaptation* transform, if *None* no chromatic adaptation is
        performed.

    Returns
    -------
    :class:`numpy.ndarray`
        Conversion matrix :math:`M`.

    Examples
    --------
    >>> from colour.models import (
    ...    RGB_COLOURSPACE_sRGB, RGB_COLOURSPACE_PROPHOTO_RGB)
    >>> matrix_RGB_to_RGB(RGB_COLOURSPACE_sRGB, RGB_COLOURSPACE_PROPHOTO_RGB)
    ... # doctest: +ELLIPSIS
    array([[ 0.5288241...,  0.3340609...,  0.1373616...],
           [ 0.0975294...,  0.8790074...,  0.0233981...],
           [ 0.0163599...,  0.1066124...,  0.8772485...]])
    """

    M = input_colourspace.matrix_RGB_to_XYZ

    if chromatic_adaptation_transform is not None:
        M_CAT = matrix_chromatic_adaptation_VonKries(
            xy_to_XYZ(input_colourspace.whitepoint),
            xy_to_XYZ(output_colourspace.whitepoint),
            chromatic_adaptation_transform,
        )

        M = matrix_dot(M_CAT, input_colourspace.matrix_RGB_to_XYZ)

    M = matrix_dot(output_colourspace.matrix_XYZ_to_RGB, M)

    return M


def RGB_to_RGB(
    RGB: ArrayLike,
    input_colourspace: RGB_Colourspace,
    output_colourspace: RGB_Colourspace,
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
    apply_cctf_decoding: Boolean = False,
    apply_cctf_encoding: Boolean = False,
    **kwargs: Any,
) -> NDArray:
    """
    Convert given *RGB* colourspace array from given input *RGB* colourspace
    to output *RGB* colourspace using given *chromatic adaptation* method.

    Parameters
    ----------
    RGB
        *RGB* colourspace array.
    input_colourspace
        *RGB* input colourspace.
    output_colourspace
        *RGB* output colourspace.
    chromatic_adaptation_transform
        *Chromatic adaptation* transform, if *None* no chromatic adaptation is
        performed.
    apply_cctf_decoding
        Apply input colourspace decoding colour component transfer function /
        electro-optical transfer function.
    apply_cctf_encoding
        Apply output colourspace encoding colour component transfer function /
        opto-electronic transfer function.

    Other Parameters
    ----------------
    kwargs
        Keywords arguments for the colour component transfer functions.

    Returns
    -------
    :class:`numpy.ndarray`
        *RGB* colourspace array.

    Notes
    -----
    +--------------------+-----------------------+---------------+
    | **Domain**         | **Scale - Reference** | **Scale - 1** |
    +====================+=======================+===============+
    | ``RGB``            | [0, 1]                | [0, 1]        |
    +--------------------+-----------------------+---------------+

    +--------------------+-----------------------+---------------+
    | **Range**          | **Scale - Reference** | **Scale - 1** |
    +====================+=======================+===============+
    | ``RGB``            | [0, 1]                | [0, 1]        |
    +--------------------+-----------------------+---------------+

    Examples
    --------
    >>> from colour.models import (
    ...     RGB_COLOURSPACE_sRGB, RGB_COLOURSPACE_PROPHOTO_RGB)
    >>> RGB = np.array([0.45595571, 0.03039702, 0.04087245])
    >>> RGB_to_RGB(RGB, RGB_COLOURSPACE_sRGB, RGB_COLOURSPACE_PROPHOTO_RGB)
    ... # doctest: +ELLIPSIS
    array([ 0.2568891...,  0.0721446...,  0.0465553...])
    """

    RGB = to_domain_1(RGB)

    if apply_cctf_decoding and input_colourspace.cctf_decoding is not None:
        with domain_range_scale("ignore"):
            RGB = input_colourspace.cctf_decoding(
                RGB, **filter_kwargs(input_colourspace.cctf_decoding, **kwargs)
            )

    M = matrix_RGB_to_RGB(
        input_colourspace, output_colourspace, chromatic_adaptation_transform
    )

    RGB = vector_dot(M, RGB)

    if apply_cctf_encoding and output_colourspace.cctf_encoding is not None:
        with domain_range_scale("ignore"):
            RGB = output_colourspace.cctf_encoding(
                RGB,
                **filter_kwargs(output_colourspace.cctf_encoding, **kwargs),
            )

    return from_range_1(RGB)



"""
Characterisation Plotting
=========================

Defines the characterisation plotting objects:

-   :func:`colour.plotting.plot_single_colour_checker`
-   :func:`colour.plotting.plot_multi_colour_checkers`
"""

from __future__ import annotations

import numpy as np
import matplotlib.pyplot as plt

from colour.hints import Any, Dict, Sequence, Tuple, Union
from colour.characterisation import ColourChecker
from colour.models import xyY_to_XYZ
from colour.plotting import (
    CONSTANTS_COLOUR_STYLE,
    ColourSwatch,
    XYZ_to_plotting_colourspace,
    artist,
    filter_colour_checkers,
    plot_multi_colour_swatches,
    override_style,
    render,
)
from colour.utilities import attest

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "plot_single_colour_checker",
    "plot_multi_colour_checkers",
]


@override_style(
    **{
        "axes.grid": False,
        "xtick.bottom": False,
        "ytick.left": False,
        "xtick.labelbottom": False,
        "ytick.labelleft": False,
    }
)
def plot_single_colour_checker(
    colour_checker: Union[
        ColourChecker, str
    ] = "ColorChecker24 - After November 2014",
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot given colour checker.

    Parameters
    ----------
    colour_checker
        Color checker to plot. ``colour_checker`` can be of any type or form
        supported by the
        :func:`colour.plotting.filter_colour_checkers` definition.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`,
        :func:`colour.plotting.plot_multi_colour_swatches`,
        :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> plot_single_colour_checker('ColorChecker 2005')  # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_Single_Colour_Checker.png
        :align: center
        :alt: plot_single_colour_checker
    """

    return plot_multi_colour_checkers([colour_checker], **kwargs)


@override_style(
    **{
        "axes.grid": False,
        "xtick.bottom": False,
        "ytick.left": False,
        "xtick.labelbottom": False,
        "ytick.labelleft": False,
    }
)
def plot_multi_colour_checkers(
    colour_checkers: Union[
        ColourChecker, str, Sequence[Union[ColourChecker, str]]
    ],
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot and compares given colour checkers.

    Parameters
    ----------
    colour_checkers
        Color checker to plot, count must be less than or equal to 2.
        ``colour_checkers`` elements can be of any type or form supported by
        the :func:`colour.plotting.filter_colour_checkers` definition.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`,
        :func:`colour.plotting.plot_multi_colour_swatches`,
        :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> plot_multi_colour_checkers(['ColorChecker 1976', 'ColorChecker 2005'])
    ... # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_Multi_Colour_Checkers.png
        :align: center
        :alt: plot_multi_colour_checkers
    """

    filtered_colour_checkers = list(
        filter_colour_checkers(colour_checkers).values()
    )

    attest(
        len(filtered_colour_checkers) <= 2,
        "Only two colour checkers can be compared at a time!",
    )

    _figure, axes = artist(**kwargs)

    compare_swatches = len(filtered_colour_checkers) == 2

    colour_swatches = []
    colour_checker_names = []
    for colour_checker in filtered_colour_checkers:
        colour_checker_names.append(colour_checker.name)
        for label, xyY in colour_checker.data.items():
            XYZ = xyY_to_XYZ(xyY)
            RGB = XYZ_to_plotting_colourspace(XYZ, colour_checker.illuminant)
            colour_swatches.append(
                ColourSwatch(np.clip(np.ravel(RGB), 0, 1), label.title())
            )

    if compare_swatches:
        colour_swatches = [
            swatch
            for pairs in zip(
                colour_swatches[0 : len(colour_swatches) // 2],
                colour_swatches[len(colour_swatches) // 2 :],
            )
            for swatch in pairs
        ]

    background_colour = "0.1"
    width = height = 1.0
    spacing = 0.25
    columns = 6

    settings: Dict[str, Any] = {
        "axes": axes,
        "width": width,
        "height": height,
        "spacing": spacing,
        "columns": columns,
        "direction": "-y",
        "text_kwargs": {"size": 8},
        "background_colour": background_colour,
        "compare_swatches": "Stacked" if compare_swatches else None,
    }
    settings.update(kwargs)
    settings["standalone"] = False

    plot_multi_colour_swatches(colour_swatches, **settings)

    axes.text(
        0.5,
        0.005,
        (
            f"{', '.join(colour_checker_names)} - "
            f"{CONSTANTS_COLOUR_STYLE.colour.colourspace.name} - "
            f"Colour Rendition Chart"
        ),
        transform=axes.transAxes,
        color=CONSTANTS_COLOUR_STYLE.colour.bright,
        ha="center",
        va="bottom",
        zorder=CONSTANTS_COLOUR_STYLE.zorder.midground_label,
    )

    settings.update(
        {
            "axes": axes,
            "standalone": True,
            "title": ", ".join(colour_checker_names),
        }
    )

    return render(**settings)



"""Showcases *CIE* chromaticity diagrams plotting examples."""

from colour import SDS_ILLUMINANTS
from colour.plotting import (
    colour_style,
    plot_chromaticity_diagram_CIE1931,
    plot_chromaticity_diagram_CIE1960UCS,
    plot_chromaticity_diagram_CIE1976UCS,
    plot_sds_in_chromaticity_diagram_CIE1931,
    plot_sds_in_chromaticity_diagram_CIE1960UCS,
    plot_sds_in_chromaticity_diagram_CIE1976UCS,
)
from colour.utilities import message_box

message_box('"CIE" Chromaticity Diagrams Plots')

colour_style()

message_box('Plotting "CIE 1931 Chromaticity Diagram".')
plot_chromaticity_diagram_CIE1931()

print("\n")

message_box('Plotting "CIE 1960 UCS Chromaticity Diagram".')
plot_chromaticity_diagram_CIE1960UCS()

print("\n")

message_box('Plotting "CIE 1976 UCS Chromaticity Diagram".')
plot_chromaticity_diagram_CIE1976UCS()

print("\n")

message_box(
    'Plotting "CIE Standard Illuminant A" and "CIE Standard Illuminant D65" '
    "spectral distribution chromaticity coordinates in "
    '"CIE 1931 Chromaticity Diagram".'
)
sd_A = SDS_ILLUMINANTS["A"]
sd_D65 = SDS_ILLUMINANTS["D65"]
plot_sds_in_chromaticity_diagram_CIE1931((sd_A, sd_D65))

print("\n")

message_box(
    'Plotting "CIE Standard Illuminant A" and "CIE Standard Illuminant D65" '
    "spectral distribution chromaticity coordinates in "
    '"CIE 1960 UCS Chromaticity Diagram".'
)
plot_sds_in_chromaticity_diagram_CIE1960UCS((sd_A, sd_D65))

print("\n")

message_box(
    'Plotting "CIE Standard Illuminant A" and "CIE Standard Illuminant D65" '
    "spectral distribution chromaticity coordinates in "
    '"CIE 1976 UCS Chromaticity Diagram".'
)
plot_sds_in_chromaticity_diagram_CIE1976UCS((sd_A, sd_D65))



from .cat import CHROMATIC_ADAPTATION_TRANSFORMS
from .cat import (
    CAT_XYZ_SCALING,
    CAT_VON_KRIES,
    CAT_BRADFORD,
    CAT_SHARP,
    CAT_FAIRCHILD,
    CAT_CMCCAT97,
    CAT_CMCCAT2000,
    CAT_CAT02,
    CAT_CAT02_BRILL2008,
    CAT_CAT16,
    CAT_BIANCO2010,
    CAT_PC_BIANCO2010,
)

__all__ = [
    "CHROMATIC_ADAPTATION_TRANSFORMS",
]
__all__ += [
    "CAT_XYZ_SCALING",
    "CAT_VON_KRIES",
    "CAT_BRADFORD",
    "CAT_SHARP",
    "CAT_FAIRCHILD",
    "CAT_CMCCAT97",
    "CAT_CMCCAT2000",
    "CAT_CAT02",
    "CAT_CAT02_BRILL2008",
    "CAT_CAT16",
    "CAT_BIANCO2010",
    "CAT_PC_BIANCO2010",
]



"""
Common RGB Colour Models Utilities
==================================

Defines various RGB colour models common utilities.
"""

from __future__ import annotations

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import ArrayLike, Boolean, Literal, NDArray, Union
from colour.models.rgb import RGB_COLOURSPACES, RGB_to_XYZ, XYZ_to_RGB

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "XYZ_to_sRGB",
    "sRGB_to_XYZ",
]


def XYZ_to_sRGB(
    XYZ: ArrayLike,
    illuminant: ArrayLike = CCS_ILLUMINANTS[
        "CIE 1931 2 Degree Standard Observer"
    ]["D65"],
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
    apply_cctf_encoding: Boolean = True,
) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to *sRGB* colourspace.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.
    illuminant
        Source illuminant chromaticity coordinates.
    chromatic_adaptation_transform
        *Chromatic adaptation* transform.
    apply_cctf_encoding
        Whether to apply the *sRGB* encoding colour component transfer function
        / inverse electro-optical transfer function.

    Returns
    -------
    :class:`numpy.ndarray`
        *sRGB* colour array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    Examples
    --------
    >>> import numpy as np
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_sRGB(XYZ)  # doctest: +ELLIPSIS
    array([ 0.7057393...,  0.1924826...,  0.2235416...])
    """

    sRGB = RGB_COLOURSPACES["sRGB"]

    return XYZ_to_RGB(
        XYZ,
        illuminant,
        sRGB.whitepoint,
        sRGB.matrix_XYZ_to_RGB,
        chromatic_adaptation_transform,
        sRGB.cctf_encoding if apply_cctf_encoding else None,
    )


def sRGB_to_XYZ(
    RGB: ArrayLike,
    illuminant: ArrayLike = CCS_ILLUMINANTS[
        "CIE 1931 2 Degree Standard Observer"
    ]["D65"],
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
    apply_cctf_decoding: Boolean = True,
) -> NDArray:
    """
    Convert from *sRGB* colourspace to *CIE XYZ* tristimulus values.

    Parameters
    ----------
    RGB
        *sRGB* colourspace array.
    illuminant
        Source illuminant chromaticity coordinates.
    chromatic_adaptation_transform
        *Chromatic adaptation* transform.
    apply_cctf_decoding
        Whether to apply the *sRGB* decoding colour component transfer function
        / electro-optical transfer function.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    Examples
    --------
    >>> import numpy as np
    >>> RGB = np.array([0.70573936, 0.19248266, 0.22354169])
    >>> sRGB_to_XYZ(RGB)  # doctest: +ELLIPSIS
    array([ 0.2065429...,  0.1219794...,  0.0513714...])
    """

    sRGB = RGB_COLOURSPACES["sRGB"]

    return RGB_to_XYZ(
        RGB,
        sRGB.whitepoint,
        illuminant,
        sRGB.matrix_RGB_to_XYZ,
        chromatic_adaptation_transform,
        sRGB.cctf_decoding if apply_cctf_decoding else None,
    )



#!/usr/bin/env python

from .datasets import *  # noqa
from . import datasets
from .machado2009 import (
    msds_cmfs_anomalous_trichromacy_Machado2009,
    matrix_anomalous_trichromacy_Machado2009,
    matrix_cvd_Machado2009,
)

__all__ = []
__all__ += datasets.__all__
__all__ += [
    "msds_cmfs_anomalous_trichromacy_Machado2009",
    "matrix_anomalous_trichromacy_Machado2009",
    "matrix_cvd_Machado2009",
]



"""
CIE 1960 UCS Colourspace
========================

Defines the *CIE 1960 UCS* colourspace transformations:

-   :func:`colour.XYZ_to_UCS`
-   :func:`colour.UCS_to_XYZ`
-   :func:`colour.UCS_to_uv`
-   :func:`colour.uv_to_UCS`
-   :func:`colour.UCS_uv_to_xy`
-   :func:`colour.xy_to_UCS_uv`

References
----------
-   :cite:`Wikipedia2008` : Wikipedia. (2008). CIE 1960 color space. Retrieved
    February 24, 2014, from http://en.wikipedia.org/wiki/CIE_1960_color_space
-   :cite:`Wikipedia2008c` : Wikipedia. (2008). Relation to CIE XYZ. Retrieved
    February 24, 2014, from
    http://en.wikipedia.org/wiki/CIE_1960_color_space#Relation_to_CIE_XYZ
"""

from __future__ import annotations

from colour.hints import ArrayLike, Floating, NDArray
from colour.utilities import (
    as_float_scalar,
    from_range_1,
    full,
    to_domain_1,
    tsplit,
    tstack,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "XYZ_to_UCS",
    "UCS_to_XYZ",
    "UCS_to_uv",
    "uv_to_UCS",
    "UCS_uv_to_xy",
    "xy_to_UCS_uv",
]


def XYZ_to_UCS(XYZ: ArrayLike) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to *CIE 1960 UCS* colourspace.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE 1960 UCS* colourspace array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``UVW``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Wikipedia2008c`, :cite:`Wikipedia2008`

    Examples
    --------
    >>> import numpy as np
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_UCS(XYZ)  # doctest: +ELLIPSIS
    array([ 0.1376933...,  0.1219722...,  0.1053731...])
    """

    X, Y, Z = tsplit(to_domain_1(XYZ))

    UVW = tstack([2 / 3 * X, Y, 1 / 2 * (-X + 3 * Y + Z)])

    return from_range_1(UVW)


def UCS_to_XYZ(UVW: ArrayLike) -> NDArray:
    """
    Convert from *CIE 1960 UCS* colourspace to *CIE XYZ* tristimulus values.

    Parameters
    ----------
    UVW
        *CIE 1960 UCS* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``UVW``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Wikipedia2008c`, :cite:`Wikipedia2008`

    Examples
    --------
    >>> import numpy as np
    >>> UVW = np.array([0.13769339, 0.12197225, 0.10537310])
    >>> UCS_to_XYZ(UVW)  # doctest: +ELLIPSIS
    array([ 0.2065400...,  0.1219722...,  0.0513695...])
    """

    U, V, W = tsplit(to_domain_1(UVW))

    XYZ = tstack([3 / 2 * U, V, 3 / 2 * U - (3 * V) + (2 * W)])

    return from_range_1(XYZ)


def UCS_to_uv(UVW: ArrayLike) -> NDArray:
    """
    Return the *uv* chromaticity coordinates from given *CIE 1960 UCS*
    colourspace array.

    Parameters
    ----------
    UVW
        *CIE 1960 UCS* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *uv* chromaticity coordinates.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``UVW``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Wikipedia2008c`

    Examples
    --------
    >>> import numpy as np
    >>> UVW = np.array([0.13769339, 0.12197225, 0.10537310])
    >>> UCS_to_uv(UVW)  # doctest: +ELLIPSIS
    array([ 0.3772021...,  0.3341350...])
    """

    U, V, W = tsplit(to_domain_1(UVW))

    U_V_W = U + V + W

    uv = tstack([U / U_V_W, V / U_V_W])

    return uv


def uv_to_UCS(uv: ArrayLike, V: Floating = 1) -> NDArray:
    """
    Return the *CIE 1960 UCS* colourspace array from given *uv* chromaticity
    coordinates.

    Parameters
    ----------
    uv
        *uv* chromaticity coordinates.
    V
        Optional :math:`V` *luminance* value used to construct the
        *CIE 1960 UCS* colourspace array, the default :math:`V` *luminance* is
        set to 1.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE 1960 UCS* colourspace array.

    References
    ----------
    :cite:`Wikipedia2008c`

    Examples
    --------
    >>> import numpy as np
    >>> uv = np.array([0.37720213, 0.33413508])
    >>> uv_to_UCS(uv)  # doctest: +ELLIPSIS
    array([ 1.1288911...,  1.        ,  0.8639104...])
    """

    u, v = tsplit(uv)
    V = as_float_scalar(to_domain_1(V))

    UVW = tstack([V * u / v, full(u.shape, V), -V * (u + v - 1) / v])

    return from_range_1(UVW)


def UCS_uv_to_xy(uv: ArrayLike) -> NDArray:
    """
    Return the *CIE xy* chromaticity coordinates from given *CIE 1960 UCS*
    colourspace *uv* chromaticity coordinates.

    Parameters
    ----------
    uv
        *CIE UCS uv* chromaticity coordinates.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xy* chromaticity coordinates.

    References
    ----------
    :cite:`Wikipedia2008c`

    Examples
    --------
    >>> import numpy as np
    >>> uv = np.array([0.37720213, 0.33413508])
    >>> UCS_uv_to_xy(uv)  # doctest: +ELLIPSIS
    array([ 0.5436955...,  0.3210794...])
    """

    u, v = tsplit(uv)

    d = 2 * u - 8 * v + 4
    xy = tstack([3 * u / d, 2 * v / d])

    return xy


def xy_to_UCS_uv(xy: ArrayLike) -> NDArray:
    """
    Return the *CIE 1960 UCS* colourspace *uv* chromaticity coordinates from
    given *CIE xy* chromaticity coordinates.

    Parameters
    ----------
    xy
        *CIE xy* chromaticity coordinates.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE UCS uv* chromaticity coordinates.

    References
    ----------
    :cite:`Wikipedia2008c`

    Examples
    --------
    >>> import numpy as np
    >>> xy = np.array([0.54369555, 0.32107941])
    >>> xy_to_UCS_uv(xy)  # doctest: +ELLIPSIS
    array([ 0.3772021...,  0.3341350...])
    """

    x, y = tsplit(xy)

    d = 12 * y - 2 * x + 3
    uv = tstack([4 * x / d, 6 * y / d])

    return uv



"""
Pivoted Log Encoding
====================

Defines the *Pivoted Log* encoding:

-   :func:`colour.models.log_encoding_PivotedLog`
-   :func:`colour.models.log_decoding_PivotedLog`

References
----------
-   :cite:`SonyImageworks2012a` : Sony Imageworks. (2012). make.py. Retrieved
    November 27, 2014, from
    https://github.com/imageworks/OpenColorIO-Configs/blob/master/\
nuke-default/make.py
"""

from __future__ import annotations

import numpy as np

from colour.hints import Floating, FloatingOrArrayLike, FloatingOrNDArray
from colour.utilities import as_float, from_range_1, to_domain_1

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "log_encoding_PivotedLog",
    "log_decoding_PivotedLog",
]


def log_encoding_PivotedLog(
    x: FloatingOrArrayLike,
    log_reference: Floating = 445,
    linear_reference: Floating = 0.18,
    negative_gamma: Floating = 0.6,
    density_per_code_value: Floating = 0.002,
) -> FloatingOrNDArray:
    """
    Define the *Josh Pines* style *Pivoted Log* log encoding curve /
    opto-electronic transfer function.

    Parameters
    ----------
    x
        Linear data :math:`x`.
    log_reference
        Log reference.
    linear_reference
        Linear reference.
    negative_gamma
        Negative gamma.
    density_per_code_value
        Density per code value.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Non-linear data :math:`y`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``x``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``y``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`SonyImageworks2012a`

    Examples
    --------
    >>> log_encoding_PivotedLog(0.18)  # doctest: +ELLIPSIS
    0.4349951...
    """

    x = to_domain_1(x)

    y = (
        log_reference
        + np.log10(x / linear_reference)
        / (density_per_code_value / negative_gamma)
    ) / 1023

    return as_float(from_range_1(y))


def log_decoding_PivotedLog(
    y: FloatingOrArrayLike,
    log_reference: Floating = 445,
    linear_reference: Floating = 0.18,
    negative_gamma: Floating = 0.6,
    density_per_code_value: Floating = 0.002,
) -> FloatingOrNDArray:
    """
    Define the *Josh Pines* style *Pivoted Log* log decoding curve /
    electro-optical transfer function.

    Parameters
    ----------
    y
        Non-linear data :math:`y`.
    log_reference
        Log reference.
    linear_reference
        Linear reference.
    negative_gamma
        Negative gamma.
    density_per_code_value
        Density per code value.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Linear data :math:`x`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``y``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``x``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`SonyImageworks2012a`

    Examples
    --------
    >>> log_decoding_PivotedLog(0.434995112414467)  # doctest: +ELLIPSIS
    0.1...
    """

    y = to_domain_1(y)

    x = (
        10
        ** (
            (y * 1023 - log_reference)
            * (density_per_code_value / negative_gamma)
        )
        * linear_reference
    )

    return as_float(from_range_1(x))



"""
:math:`J_za_zb_z` Colourspace
=============================

Defines the :math:`J_za_zb_z` colourspace:

-   :func:`colour.models.IZAZBZ_METHODS`
-   :func:`colour.models.XYZ_to_Izazbz`
-   :func:`colour.models.Izazbz_to_XYZ`
-   :func:`colour.XYZ_to_Jzazbz`
-   :func:`colour.Jzazbz_to_XYZ`

References
----------
-   :cite:`Safdar2017` : Safdar, M., Cui, G., Kim, Y. J., & Luo, M. R. (2017).
    Perceptually uniform color space for image signals including high dynamic
    range and wide gamut. Optics Express, 25(13), 15131.
    doi:10.1364/OE.25.015131
-   :cite:`Safdar2021` : Safdar, M., Hardeberg, J. Y., & Ronnier Luo, M.
    (2021). ZCAM, a colour appearance model based on a high dynamic range
    uniform colour space. Optics Express, 29(4), 6036. doi:10.1364/OE.413659
"""

from __future__ import annotations

import numpy as np

from colour.algebra import vector_dot
from colour.hints import (
    ArrayLike,
    Literal,
    NDArray,
    Optional,
    Tuple,
    Union,
)
from colour.models.rgb.transfer_functions import (
    eotf_inverse_ST2084,
    eotf_ST2084,
)
from colour.models.rgb.transfer_functions.st_2084 import CONSTANTS_ST2084
from colour.utilities import (
    Structure,
    as_float_array,
    domain_range_scale,
    optional,
    tsplit,
    tstack,
    validate_method,
)
from colour.utilities.documentation import (
    DocstringTuple,
    is_documentation_building,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CONSTANTS_JZAZBZ_SAFDAR2017",
    "CONSTANTS_JZAZBZ_SAFDAR2021",
    "MATRIX_JZAZBZ_XYZ_TO_LMS",
    "MATRIX_JZAZBZ_LMS_TO_XYZ",
    "MATRIX_JZAZBZ_LMS_P_TO_IZAZBZ_SAFDAR2017",
    "MATRIX_JZAZBZ_IZAZBZ_TO_LMS_P_SAFDAR2017",
    "MATRIX_JZAZBZ_LMS_P_TO_IZAZBZ_SAFDAR2021",
    "MATRIX_JZAZBZ_IZAZBZ_TO_LMS_P_SAFDAR2021",
    "IZAZBZ_METHODS",
    "XYZ_to_Izazbz",
    "Izazbz_to_XYZ",
    "XYZ_to_Jzazbz",
    "Jzazbz_to_XYZ",
]

CONSTANTS_JZAZBZ_SAFDAR2017: Structure = Structure(
    b=1.15, g=0.66, d=-0.56, d_0=1.6295499532821566 * 10**-11
)
CONSTANTS_JZAZBZ_SAFDAR2017.update(CONSTANTS_ST2084)
CONSTANTS_JZAZBZ_SAFDAR2017.m_2 = 1.7 * 2523 / 2**5
"""
Constants for :math:`J_za_zb_z` colourspace and its variant of the perceptual
quantizer (PQ) from Dolby Laboratories.

Notes
-----
-   The :math:`m2` constant, i.e. the power factor has been re-optimized during
    the development of the :math:`J_za_zb_z` colourspace.
"""

CONSTANTS_JZAZBZ_SAFDAR2021: Structure = Structure(
    **CONSTANTS_JZAZBZ_SAFDAR2017
)
CONSTANTS_JZAZBZ_SAFDAR2021.d_0 = 3.7035226210190005 * 10**-11
""":math:`J_za_zb_z` colourspace constants for the *ZCAM* colour appearance model."""

MATRIX_JZAZBZ_XYZ_TO_LMS: NDArray = np.array(
    [
        [0.41478972, 0.579999, 0.0146480],
        [-0.2015100, 1.120649, 0.0531008],
        [-0.0166008, 0.264800, 0.6684799],
    ]
)
"""
:math:`J_za_zb_z` *CIE XYZ* tristimulus values to normalised cone responses
matrix.
"""

MATRIX_JZAZBZ_LMS_TO_XYZ: NDArray = np.linalg.inv(MATRIX_JZAZBZ_XYZ_TO_LMS)
"""
:math:`J_za_zb_z` normalised cone responses to *CIE XYZ* tristimulus values
matrix.
"""

MATRIX_JZAZBZ_LMS_P_TO_IZAZBZ_SAFDAR2017: NDArray = np.array(
    [
        [0.500000, 0.500000, 0.000000],
        [3.524000, -4.066708, 0.542708],
        [0.199076, 1.096799, -1.295875],
    ]
)
"""
:math:`LMS_p` *SMPTE ST 2084:2014* encoded normalised cone responses to
:math:`I_za_zb_z` intermediate colourspace matrix.
"""

MATRIX_JZAZBZ_IZAZBZ_TO_LMS_P_SAFDAR2017: NDArray = np.linalg.inv(
    MATRIX_JZAZBZ_LMS_P_TO_IZAZBZ_SAFDAR2017
)
"""
:math:`I_za_zb_z` intermediate colourspace to :math:`LMS_p`
*SMPTE ST 2084:2014* encoded normalised cone responses matrix.
"""

MATRIX_JZAZBZ_LMS_P_TO_IZAZBZ_SAFDAR2021: NDArray = np.array(
    [
        [0.000000, 1.000000, 0.000000],
        [3.524000, -4.066708, 0.542708],
        [0.199076, 1.096799, -1.295875],
    ]
)
"""
:math:`LMS_p` *SMPTE ST 2084:2014* encoded normalised cone responses to
:math:`I_za_zb_z` intermediate colourspace matrix.

References
----------
:cite:`Safdar2021`
"""

MATRIX_JZAZBZ_IZAZBZ_TO_LMS_P_SAFDAR2021: NDArray = np.linalg.inv(
    MATRIX_JZAZBZ_LMS_P_TO_IZAZBZ_SAFDAR2021
)
"""
:math:`I_za_zb_z` intermediate colourspace to :math:`LMS_p`
*SMPTE ST 2084:2014* encoded normalised cone responses matrix.

References
----------
:cite:`Safdar2021`
"""

IZAZBZ_METHODS: Tuple = ("Safdar 2017", "Safdar 2021", "ZCAM")
if is_documentation_building():  # pragma: no cover
    IZAZBZ_METHODS = DocstringTuple(IZAZBZ_METHODS)
    IZAZBZ_METHODS.__doc__ = """
Supported :math:`I_za_zb_z` computation methods.

References
----------
:cite:`Safdar2017`, :cite:`Safdar2021`
"""


def XYZ_to_Izazbz(
    XYZ_D65: ArrayLike,
    constants: Optional[Structure] = None,
    method: Union[
        Literal["Safdar 2017", "Safdar 2021", "ZCAM"], str
    ] = "Safdar 2017",
) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to :math:`I_za_zb_z`
    colourspace.

    Parameters
    ----------
    XYZ_D65
        *CIE XYZ* tristimulus values under
        *CIE Standard Illuminant D Series D65*.
    constants
        :math:`J_za_zb_z` colourspace constants.
    method
        Computation methods, *Safdar 2021* and *ZCAM* methods are equivalent.

    Returns
    -------
    :class:`numpy.ndarray`
        :math:`I_za_zb_z` colourspace array where :math:`I_z` is the achromatic
        response, :math:`a_z` is redness-greenness and :math:`b_z` is
        yellowness-blueness.

    Warnings
    --------
    The underlying *SMPTE ST 2084:2014* transfer function is an absolute
    transfer function.

    Notes
    -----
    -   The underlying *SMPTE ST 2084:2014* transfer function is an absolute
        transfer function, thus the domain and range values for the *Reference*
        and *1* scales are only indicative that the data is not affected by
        scale transformations. The effective domain of *SMPTE ST 2084:2014*
        inverse electro-optical transfer function (EOTF) is
        [0.0001, 10000].

    +------------+-----------------------+------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``XYZ``    | ``UN``                | ``UN``           |
    +------------+-----------------------+------------------+

    +------------+-----------------------+------------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``Izazbz`` | ``Iz`` : [0, 1]       | ``Iz`` : [0, 1]  |
    |            |                       |                  |
    |            | ``az`` : [-1, 1]      | ``az`` : [-1, 1] |
    |            |                       |                  |
    |            | ``bz`` : [-1, 1]      | ``bz`` : [-1, 1] |
    +------------+-----------------------+------------------+

    References
    ----------
    :cite:`Safdar2017`, :cite:`Safdar2021`

    Examples
    --------
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_Izazbz(XYZ)  # doctest: +ELLIPSIS
    array([ 0.0120779...,  0.0092430...,  0.0052600...])
    """

    X_D65, Y_D65, Z_D65 = tsplit(as_float_array(XYZ_D65))

    method = validate_method(method, IZAZBZ_METHODS)

    constants = optional(
        constants,
        CONSTANTS_JZAZBZ_SAFDAR2017
        if method == "safdar 2017"
        else CONSTANTS_JZAZBZ_SAFDAR2021,
    )

    X_p_D65 = constants.b * X_D65 - (constants.b - 1) * Z_D65
    Y_p_D65 = constants.g * Y_D65 - (constants.g - 1) * X_D65

    XYZ_p_D65 = tstack([X_p_D65, Y_p_D65, Z_D65])

    LMS = vector_dot(MATRIX_JZAZBZ_XYZ_TO_LMS, XYZ_p_D65)

    with domain_range_scale("ignore"):
        LMS_p = eotf_inverse_ST2084(LMS, 10000, constants)

    if method == "safdar 2017":
        Izazbz = vector_dot(MATRIX_JZAZBZ_LMS_P_TO_IZAZBZ_SAFDAR2017, LMS_p)
    else:
        Izazbz = vector_dot(MATRIX_JZAZBZ_LMS_P_TO_IZAZBZ_SAFDAR2021, LMS_p)
        Izazbz[..., 0] -= constants.d_0

    return Izazbz


def Izazbz_to_XYZ(
    Izazbz: ArrayLike,
    constants: Optional[Structure] = None,
    method: Union[
        Literal["Safdar 2017", "Safdar 2021", "ZCAM"], str
    ] = "Safdar 2017",
) -> NDArray:
    """
    Convert from :math:`I_za_zb_z` colourspace to *CIE XYZ* tristimulus
    values.

    Parameters
    ----------
    Izazbz
        :math:`I_za_zb_z` colourspace array where :math:`I_z` is the
        achromatic response, :math:`a_z` is redness-greenness and
        :math:`b_z` is yellowness-blueness.
    constants
        :math:`J_za_zb_z` colourspace constants.
    method
        Computation methods, *Safdar 2021* and *ZCAM* methods are equivalent.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values under
        *CIE Standard Illuminant D Series D65*.

    Warnings
    --------
    The underlying *SMPTE ST 2084:2014* transfer function is an absolute
    transfer function.

    Notes
    -----
    -   The underlying *SMPTE ST 2084:2014* transfer function is an absolute
        transfer function, thus the domain and range values for the *Reference*
        and *1* scales are only indicative that the data is not affected by
        scale transformations.

    +------------+-----------------------+------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``Izazbz`` | ``Iz`` : [0, 1]       | ``Iz`` : [0, 1]  |
    |            |                       |                  |
    |            | ``az`` : [-1, 1]      | ``az`` : [-1, 1] |
    |            |                       |                  |
    |            | ``bz`` : [-1, 1]      | ``bz`` : [-1, 1] |
    +------------+-----------------------+------------------+

    +------------+-----------------------+------------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``XYZ``    | ``UN``                | ``UN``           |
    +------------+-----------------------+------------------+

    References
    ----------
    :cite:`Safdar2017`, :cite:`Safdar2021`

    Examples
    --------
    >>> Izazbz = np.array([0.01207793, 0.00924302, 0.00526007])
    >>> Izazbz_to_XYZ(Izazbz)  # doctest: +ELLIPSIS
    array([ 0.2065401...,  0.1219723...,  0.0513696...])
    """

    Izazbz = as_float_array(Izazbz)

    method = validate_method(method, IZAZBZ_METHODS)

    constants = optional(
        constants,
        CONSTANTS_JZAZBZ_SAFDAR2017
        if method == "safdar 2017"
        else CONSTANTS_JZAZBZ_SAFDAR2021,
    )

    if method == "safdar 2017":
        LMS_p = vector_dot(MATRIX_JZAZBZ_IZAZBZ_TO_LMS_P_SAFDAR2017, Izazbz)
    else:
        Izazbz[..., 0] += constants.d_0
        LMS_p = vector_dot(MATRIX_JZAZBZ_IZAZBZ_TO_LMS_P_SAFDAR2021, Izazbz)

    with domain_range_scale("ignore"):
        LMS = eotf_ST2084(LMS_p, 10000, constants)

    X_p_D65, Y_p_D65, Z_p_D65 = tsplit(
        vector_dot(MATRIX_JZAZBZ_LMS_TO_XYZ, LMS)
    )

    X_D65 = (X_p_D65 + (constants.b - 1) * Z_p_D65) / constants.b
    Y_D65 = (Y_p_D65 + (constants.g - 1) * X_D65) / constants.g

    XYZ_D65 = tstack([X_D65, Y_D65, Z_p_D65])

    return XYZ_D65


def XYZ_to_Jzazbz(
    XYZ_D65: ArrayLike, constants: Structure = CONSTANTS_JZAZBZ_SAFDAR2017
) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to :math:`J_za_zb_z`
    colourspace.

    Parameters
    ----------
    XYZ_D65
        *CIE XYZ* tristimulus values under
        *CIE Standard Illuminant D Series D65*.
    constants
        :math:`J_za_zb_z` colourspace constants.

    Returns
    -------
    :class:`numpy.ndarray`
        :math:`J_za_zb_z` colourspace array where :math:`J_z` is Lightness,
        :math:`a_z` is redness-greenness and :math:`b_z` is
        yellowness-blueness.

    Warnings
    --------
    The underlying *SMPTE ST 2084:2014* transfer function is an absolute
    transfer function.

    Notes
    -----
    -   The underlying *SMPTE ST 2084:2014* transfer function is an absolute
        transfer function, thus the domain and range values for the *Reference*
        and *1* scales are only indicative that the data is not affected by
        scale transformations. The effective domain of *SMPTE ST 2084:2014*
        inverse electro-optical transfer function (EOTF) is
        [0.0001, 10000].

    +------------+-----------------------+------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``XYZ``    | ``UN``                | ``UN``           |
    +------------+-----------------------+------------------+

    +------------+-----------------------+------------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``Jzazbz`` | ``Jz`` : [0, 1]       | ``Jz`` : [0, 1]  |
    |            |                       |                  |
    |            | ``az`` : [-1, 1]      | ``az`` : [-1, 1] |
    |            |                       |                  |
    |            | ``bz`` : [-1, 1]      | ``bz`` : [-1, 1] |
    +------------+-----------------------+------------------+

    References
    ----------
    :cite:`Safdar2017`

    Examples
    --------
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_Jzazbz(XYZ)  # doctest: +ELLIPSIS
    array([ 0.0053504...,  0.0092430...,  0.0052600...])
    """

    XYZ_D65 = as_float_array(XYZ_D65)

    with domain_range_scale("ignore"):
        I_z, a_z, b_z = tsplit(
            XYZ_to_Izazbz(XYZ_D65, CONSTANTS_JZAZBZ_SAFDAR2017, "Safdar 2017")
        )

    J_z = ((1 + constants.d) * I_z) / (1 + constants.d * I_z) - constants.d_0

    Jzazbz = tstack([J_z, a_z, b_z])

    return Jzazbz


def Jzazbz_to_XYZ(
    Jzazbz: ArrayLike, constants: Structure = CONSTANTS_JZAZBZ_SAFDAR2017
) -> NDArray:
    """
    Convert from :math:`J_za_zb_z` colourspace to *CIE XYZ* tristimulus
    values.

    Parameters
    ----------
    Jzazbz
        :math:`J_za_zb_z` colourspace array  where :math:`J_z` is Lightness,
        :math:`a_z` is redness-greenness and :math:`b_z` is
        yellowness-blueness.
    constants
        :math:`J_za_zb_z` colourspace constants.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values under
        *CIE Standard Illuminant D Series D65*.

    Warnings
    --------
    The underlying *SMPTE ST 2084:2014* transfer function is an absolute
    transfer function.

    Notes
    -----
    -   The underlying *SMPTE ST 2084:2014* transfer function is an absolute
        transfer function, thus the domain and range values for the *Reference*
        and *1* scales are only indicative that the data is not affected by
        scale transformations.

    +------------+-----------------------+------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``Jzazbz`` | ``Jz`` : [0, 1]       | ``Jz`` : [0, 1]  |
    |            |                       |                  |
    |            | ``az`` : [-1, 1]      | ``az`` : [-1, 1] |
    |            |                       |                  |
    |            | ``bz`` : [-1, 1]      | ``bz`` : [-1, 1] |
    +------------+-----------------------+------------------+

    +------------+-----------------------+------------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``XYZ``    | ``UN``                | ``UN``           |
    +------------+-----------------------+------------------+

    References
    ----------
    :cite:`Safdar2017`

    Examples
    --------
    >>> Jzazbz = np.array([0.00535048, 0.00924302, 0.00526007])
    >>> Jzazbz_to_XYZ(Jzazbz)  # doctest: +ELLIPSIS
    array([ 0.2065402...,  0.1219723...,  0.0513696...])
    """

    J_z, a_z, b_z = tsplit(as_float_array(Jzazbz))

    I_z = (J_z + constants.d_0) / (
        1 + constants.d - constants.d * (J_z + constants.d_0)
    )

    with domain_range_scale("ignore"):
        XYZ_D65 = Izazbz_to_XYZ(
            tstack([I_z, a_z, b_z]), CONSTANTS_JZAZBZ_SAFDAR2017, "Safdar 2017"
        )

    return XYZ_D65



"""Showcases gamut section plotting examples."""

import numpy as np
from matplotlib.lines import Line2D

import colour.plotting
from colour.plotting import (
    colour_style,
    plot_visible_spectrum_section,
    plot_RGB_colourspace_section,
)
from colour.utilities import message_box

message_box("Gamut Section Plots")

colour_style()

message_box(
    'Plotting a "Visible Spectrum" section at 50% "Lightness" in the '
    '"CIE Luv" colourspace.'
)

plot_visible_spectrum_section(
    model="CIE Luv",
    origin=0.5,
)

print("\n")

message_box(
    'Plotting a "Visible Spectrum" section at 50% "Lightness" in the '
    '"CIE Luv" colourspace and customising the section styling.'
)

plot_visible_spectrum_section(
    model="CIE Luv", origin=0.5, section_colours="RGB", section_opacity=0.15
)

print("\n")

message_box(
    'Plotting a "Visible Spectrum" section at 50% "Lightness" in the '
    '"CIE Luv" colourspace.'
)

plot_visible_spectrum_section(model="CIE Luv", origin=0.5)

print("\n")

message_box(
    'Plotting a "Visible Spectrum" section at 25% along the "u" axis in the '
    '"CIE Luv" colourspace.'
)

plot_visible_spectrum_section(
    model="CIE Luv",
    axis="+x",
    origin=0.25,
    section_colours="RGB",
    section_opacity=0.15,
)

print("\n")

message_box(
    'Plotting a "sRGB" colourspace section at 50% "Lightness" in the '
    '"ICtCp" colourspace using section normalisation.'
)

plot_RGB_colourspace_section(
    colourspace="sRGB", model="ICtCp", origin=0.5, normalise=True
)

print("\n")

message_box(
    'Combining multiple hull sections together at 25% "Lightness" in the '
    '"Oklab" colourspace.'
)

figure, axes = plot_visible_spectrum_section(
    model="Oklab", origin=0.25, section_opacity=0.15, standalone=False
)
plot_RGB_colourspace_section(
    colourspace="sRGB",
    model="Oklab",
    origin=0.25,
    section_colours="RGB",
    section_opacity=0.15,
    contour_colours="RGB",
    axes=axes,
)

print("\n")

message_box(
    'Combining multiple hull sections together at varying "Lightness" in the '
    '"DIN99" colourspace.'
)

figure, axes = plot_visible_spectrum_section(
    model="DIN99", origin=0.5, section_opacity=0.15, standalone=False
)

bounding_box = [
    axes.get_xlim()[0],
    axes.get_xlim()[1],
    axes.get_ylim()[0],
    axes.get_ylim()[1],
]

section_colours = colour.notation.HEX_to_RGB(
    colour.plotting.CONSTANTS_COLOUR_STYLE.colour.cycle[:4]
)

origins = []
legend_lines = []
for i, RGB in zip(np.arange(0.5, 0.9, 0.1), section_colours):
    origins.append(i * 100)
    plot_RGB_colourspace_section(
        colourspace="sRGB",
        model="DIN99",
        origin=i,
        section_colours=RGB,
        section_opacity=0.15,
        contour_colours=RGB,
        axes=axes,
        standalone=False,
    )
    legend_lines.append(Line2D([0], [0], color=RGB, label=f"{i * 100}%"))

axes.legend(handles=legend_lines)

colour.plotting.render(
    title=f"Visible Spectrum - 50% - sRGB Sections - {origins}% -  DIN99",
    axes=axes,
    bounding_box=bounding_box,
)



"""Showcases overall *Colour* examples."""

import numpy as np
import warnings

import colour
from colour.utilities import (
    filter_warnings,
    message_box,
    warning,
    runtime_warning,
    usage_warning,
)

message_box("Automatic Colour Conversion Graph")

message_box(
    'Starting with version "0.3.14", "Colour" implements an automatic colour '
    "conversion graph enabling easier colour conversions."
)

message_box(
    'Converting a "ColorChecker" "dark skin" sample spectral distribution to '
    '"Output-Referred" "sRGB" colourspace.'
)

sd = colour.SDS_COLOURCHECKERS["ColorChecker N Ohta"]["dark skin"]

print(colour.convert(sd, "Spectral Distribution", "sRGB"))

print("\n")

RGB = np.array([0.45675795, 0.30986982, 0.24861924])
message_box(
    f'Converting to the "CAM16-UCS" colourspace from given "Output-Referred" '
    f'"sRGB" colourspace values:\n\n\t{RGB}'
)
print(colour.convert(RGB, "Output-Referred RGB", "CAM16UCS"))

print("\n")

Jpapbp = np.array([0.39994811, 0.09206558, 0.0812752])
message_box(
    f'Converting to the "Output-Referred" "sRGB" colourspace from given '
    f'"CAM16-UCS" colourspace colourspace values:\n\n\t{RGB}'
)
print(colour.convert(Jpapbp, "CAM16UCS", "sRGB"))

print("\n")

message_box('Filter "Colour" Warnings')

warning("This is a first warning and it can be filtered!")

filter_warnings()

warning("This is a second warning and it has been filtered!")

filter_warnings(False)

warning("This is a third warning and it has not been filtered!")

message_box(
    "All Python can be filtered by setting the "
    '"colour.utilities.filter_warnings" definition "python_warnings" '
    "argument."
)

warnings.warn("This is a fourth warning and it has not been filtered!")

filter_warnings(python_warnings=False)

warning("This is a fifth warning and it has been filtered!")

filter_warnings(False, python_warnings=False)

warning("This is a sixth warning and it has not been filtered!")

filter_warnings(False, python_warnings=False)

filter_warnings(colour_warnings=False, colour_runtime_warnings=True)

runtime_warning("This is a first runtime warning and it has been filtered!")

filter_warnings(colour_warnings=False, colour_usage_warnings=True)

usage_warning("This is a first usage warning and it has been filtered!")

print("\n")

message_box('Overall "Colour" Examples')

message_box("N-Dimensional Arrays Support")

XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
illuminant = np.array([0.31270, 0.32900])
message_box(f'Using 1d "ArrayLike" parameter:\n\n{XYZ}')
print(colour.XYZ_to_Lab(XYZ, illuminant=illuminant))

print("\n")

XYZ = np.tile(XYZ, (6, 1))
illuminant = np.tile(illuminant, (6, 1))
message_box(f'Using 2d "ArrayLike" parameter:\n\n{XYZ}')
print(colour.XYZ_to_Lab(XYZ, illuminant=illuminant))

print("\n")

XYZ = np.reshape(XYZ, (2, 3, 3))
illuminant = np.reshape(illuminant, (2, 3, 2))
message_box(f'Using 3d "ArrayLike" parameter:\n\n{XYZ}')
print(colour.XYZ_to_Lab(XYZ, illuminant=illuminant))

print("\n")

XYZ = np.reshape(XYZ, (3, 2, 1, 3))
illuminant = np.reshape(illuminant, (3, 2, 1, 2))
message_box(f'Using 4d "ArrayLike" parameter:\n\n{XYZ}')
print(colour.XYZ_to_Lab(XYZ, illuminant=illuminant))

print("\n")

xy = np.tile((0.31270, 0.32900), (6, 1))
message_box(
    f"Definitions return value may lose a dimension with respect to the "
    f"parameter(s):\n\n{xy}"
)
print(colour.xy_to_CCT(xy))

print("\n")

CCT = np.tile(6504.38938305, 6)
message_box(
    f"Definitions return value may gain a dimension with respect to the "
    f"parameter(s):\n\n{CCT}"
)
print(colour.CCT_to_xy(CCT))

print("\n")

message_box(
    'Definitions mixing "ArrayLike" and "Number" parameters expect the '
    '"Number" parameters to have a dimension less than the "ArrayLike" '
    "parameters."
)
XYZ_1 = np.array([28.00, 21.26, 5.27])
xy_o1 = np.array([0.4476, 0.4074])
xy_o2 = np.array([0.3127, 0.3290])
Y_o = 20
E_o1 = 1000
E_o2 = 1000
message_box(
    f"Parameters:\n\n"
    f"\tXYZ_1: {XYZ_1}\n"
    f"\txy_o1: {xy_o1}\n"
    f"\txy_o2: {xy_o2}\n"
    f"\tY_o: {Y_o}\n"
    f"\tE_o1: {E_o1}\n"
    f"\tE_o2: {E_o2}"
)
print(
    colour.adaptation.chromatic_adaptation_CIE1994(
        XYZ_1, xy_o1, xy_o2, Y_o, E_o1, E_o2
    )
)

print("\n")

XYZ_1 = np.tile(XYZ_1, (6, 1))
message_box(
    f"Parameters:\n\n"
    f"\tXYZ_1: {XYZ_1}\n"
    f"\txy_o1: {xy_o1}\n"
    f"\txy_o2: {xy_o2}\n"
    f"\tY_o: {Y_o}\n"
    f"\tE_o1: {E_o1}\n"
    f"\tE_o2: {E_o2}"
)
print(
    colour.adaptation.chromatic_adaptation_CIE1994(
        XYZ_1, xy_o1, xy_o2, Y_o, E_o1, E_o2
    )
)

print("\n")

xy_o1 = np.tile(xy_o1, (6, 1))
xy_o2 = np.tile(xy_o2, (6, 1))
Y_o = np.tile(Y_o, 6)
E_o1 = np.tile(E_o1, 6)
E_o2 = np.tile(E_o2, 6)
message_box(
    f"Parameters:\n\n"
    f"\tXYZ_1: {XYZ_1}\n"
    f"\txy_o1: {xy_o1}\n"
    f"\txy_o2: {xy_o2}\n"
    f"\tY_o: {Y_o}\n"
    f"\tE_o1: {E_o1}\n"
    f"\tE_o2: {E_o2}"
)
print(
    colour.adaptation.chromatic_adaptation_CIE1994(
        XYZ_1, xy_o1, xy_o2, Y_o, E_o1, E_o2
    )
)

print("\n")

XYZ_1 = np.reshape(XYZ_1, (2, 3, 3))
xy_o1 = np.reshape(xy_o1, (2, 3, 2))
xy_o2 = np.reshape(xy_o2, (2, 3, 2))
Y_o = np.reshape(Y_o, (2, 3))
E_o1 = np.reshape(E_o1, (2, 3))
E_o2 = np.reshape(E_o2, (2, 3))
message_box(
    f"Parameters:\n\n"
    f"\tXYZ_1: {XYZ_1}\n"
    f"\txy_o1: {xy_o1}\n"
    f"\txy_o2: {xy_o2}\n"
    f"\tY_o: {Y_o}\n"
    f"\tE_o1: {E_o1}\n"
    f"\tE_o2: {E_o2}"
)
print(
    colour.adaptation.chromatic_adaptation_CIE1994(
        XYZ_1, xy_o1, xy_o2, Y_o, E_o1, E_o2
    )
)

print("\n")

message_box("Domain-Range Scales")

message_box(
    '"Colour" uses two different domain-range scales: \n\n'
    '- "Reference"\n'
    '- "1"'
)

print("\n")

message_box("Printing the current domain-range scale:")

print(colour.get_domain_range_scale())

print("\n")

message_box('Setting the current domain-range scale to "1":')

colour.set_domain_range_scale("1")

XYZ_1 = np.array([0.2800, 0.2126, 0.0527])
xy_o1 = np.array([0.4476, 0.4074])
xy_o2 = np.array([0.3127, 0.3290])
Y_o = 0.2
E_o1 = 1000
E_o2 = 1000
message_box(
    f"Parameters:\n\n"
    f"\tXYZ_1: {XYZ_1}\n"
    f"\txy_o1: {xy_o1}\n"
    f"\txy_o2: {xy_o2}\n"
    f"\tY_o: {Y_o}\n"
    f"\tE_o1: {E_o1}\n"
    f"\tE_o2: {E_o2}"
)
print(
    colour.adaptation.chromatic_adaptation_CIE1994(
        XYZ_1, xy_o1, xy_o2, Y_o, E_o1, E_o2
    )
)



"""
Blackmagic Design Transfer Functions
====================================

Defines the *Blackmagic Design* colour component transfer functions:

-   :func:`colour.models.oetf_BlackmagicFilmGeneration5`
-   :func:`colour.models.oetf_inverse_BlackmagicFilmGeneration5`

References
----------
-   :cite:`BlackmagicDesign2021` : Blackmagic Design. (2021). Blackmagic
    Generation 5 Color Science. https://drive.google.com/file/d/\
1FF5WO2nvI9GEWb4_EntrBoV9ZIuFToZd/view
"""

from __future__ import annotations

import numpy as np

from colour.hints import FloatingOrArrayLike, FloatingOrNDArray
from colour.utilities import Structure, as_float, from_range_1, to_domain_1

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CONSTANTS_BLACKMAGIC_FILM_GENERATION_5",
    "oetf_BlackmagicFilmGeneration5",
    "oetf_inverse_BlackmagicFilmGeneration5",
]

CONSTANTS_BLACKMAGIC_FILM_GENERATION_5: Structure = Structure(
    A=0.08692876065491224,
    B=0.005494072432257808,
    C=0.5300133392291939,
    D=8.283605932402494,
    E=0.09246575342465753,
    LIN_CUT=0.005,
)
"""*Blackmagic Film Generation 5* colour component transfer functions constants."""


def oetf_BlackmagicFilmGeneration5(
    x: FloatingOrArrayLike,
    constants: Structure = CONSTANTS_BLACKMAGIC_FILM_GENERATION_5,
) -> FloatingOrNDArray:
    """
    Define the *Blackmagic Film Generation 5* opto-electronic transfer
    function.

    Parameters
    ----------
    x
        Linear light value :math`x`.
    constants
        *Blackmagic Film Generation 5* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Encoded value :math:`y`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``x``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``y``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`BlackmagicDesign2021`

    Examples
    --------
    >>> oetf_BlackmagicFilmGeneration5(0.18)  # doctest: +ELLIPSIS
    0.3835616...
    """

    x = to_domain_1(x)

    A = constants.A
    B = constants.B
    C = constants.C
    D = constants.D
    E = constants.E
    LIN_CUT = constants.LIN_CUT

    V_out = np.where(
        x < LIN_CUT,
        D * x + E,
        A * np.log(x + B) + C,
    )

    return as_float(from_range_1(V_out))


def oetf_inverse_BlackmagicFilmGeneration5(
    y: FloatingOrArrayLike,
    constants: Structure = CONSTANTS_BLACKMAGIC_FILM_GENERATION_5,
) -> FloatingOrNDArray:
    """
    Define the *Blackmagic Film Generation 5* inverse opto-electronic transfer
    function (OETF).

    Parameters
    ----------
    y
        Encoded value :math:`y`.
    constants
        *Blackmagic Film Generation 5* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Linear light value :math`x`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``y``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``x``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`BlackmagicDesign2021`

    Examples
    --------
    >>> oetf_inverse_BlackmagicFilmGeneration5(0.38356164383561653)
    ... # doctest: +ELLIPSIS
    0.1799999...
    """

    y = to_domain_1(y)

    A = constants.A
    B = constants.B
    C = constants.C
    D = constants.D
    E = constants.E
    LIN_CUT = constants.LIN_CUT

    LOG_CUT = D * LIN_CUT + E

    x = np.where(
        y < LOG_CUT,
        (y - E) / D,
        np.exp((y - C) / A) - B,
    )
    return as_float(from_range_1(x))
